"""
=================================================
Earthquake (:mod:`geoana.earthquake.oksar`)
=================================================
.. currentmodule:: geoana.earthquake.oksar

The ``geoana.earthquake.oksar`` etc...



.. autosummary::
  :toctree: generated/

  EarthquakeInterferogram
  Oksar

"""
from . import oksar



"""
=================================================
Base Electromagnetics (:mod:`geoana.em`)
=================================================
.. currentmodule:: geoana.em

The ``geoana.em.base`` module contains base classes whose properties are relevant
to a large range of electromagnetic problems. These classes are not used directly
to compute solutions to electromagnetic problems.


Base Classes
============
.. autosummary::
  :toctree: generated/

  BaseEM
  BaseDipole
  BaseElectricDipole
  BaseMagneticDipole
  BaseLineCurrent

"""
from . import static
from . import fdem
from . import tdem
from .base import BaseEM, BaseDipole, BaseElectricDipole, BaseMagneticDipole, BaseLineCurrent





"""
======================================================
Gravity (:mod:`geoana.gravity`)
======================================================
.. currentmodule:: geoana.gravity

The ``geoana.gravity`` module contains simulation classes for solving
basic gravitational problems.

Simulation Classes
==================
.. autosummary::
  :toctree: generated/

  PointMass
  Sphere
"""

import numpy as np
from scipy.constants import G


class PointMass:
    """Class for gravitational solutions for a point mass.

    The ``PointMass`` class is used to analytically compute the gravitational
    potentials, fields, and gradients for a point mass.

    Parameters
    ----------
    mass : float
        Mass of the point particle (kg). Default is m = 1 kg
    location : array_like, optional
        Location of the point mass in 3D space (m). Default is (0, 0, 0)
    """

    def __init__(self, mass=1.0, location=None, **kwargs):

        self.mass = mass
        if location is None:
            location = np.r_[0, 0, 0]
        self.location = location
        super().__init__(**kwargs)

    @property
    def mass(self):
        """Mass of the point particle in kilograms.

        Returns
        -------
        float
            Mass of the point particle in kilograms.
        """
        return self._mass

    @mass.setter
    def mass(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"mass must be a number, got {type(value)}")

        self._mass = value

    @property
    def location(self):
        """Location of the point mass.

        Returns
        -------
        (3) numpy.ndarray of float
            Location of the point mass in meters.  Default = np.r_[0,0,0]
        """
        return self._location

    @location.setter
    def location(self, vec):

        try:
            vec = np.asarray(vec, dtype=float)
        except:
            raise TypeError(f"location must be array_like of float, got {type(vec)}")

        vec = np.squeeze(vec)
        if vec.shape != (3,):
            raise ValueError(
                f"location must be array_like with shape (3,), got {vec.shape}"
            )

        self._location = vec

    def gravitational_potential(self, xyz):
        """
        Gravitational potential due to a point mass.  See Blakely, 1996
        equation 3.4.

        .. math::

            U(P) = \\gamma \\frac{m}{r}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Point mass location in units m.

        Returns
        -------
        (..., ) numpy.ndarray
            Gravitational potential at point mass location xyz in units :math:`\\frac{m^2}{s^2}`.

        Examples
        --------
        Here, we define a point mass with mass=1kg and plot the gravitational
        potential as a function of distance.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from geoana.gravity import PointMass

        Define the point mass.

        >>> location = np.r_[0., 0., 0.]
        >>> mass = 1.0
        >>> simulation = PointMass(
        >>>     mass=mass, location=location
        >>> )

        Now we create a set of gridded locations, take the distances and compute the gravitational potential.

        >>> X, Y = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> r = np.linalg.norm(xyz, axis=-1)
        >>> u = simulation.gravitational_potential(xyz)

        Finally, we plot the gravitational potential as a function of distance.

        >>> plt.plot(r, u)
        >>> plt.xlabel('Distance from point mass')
        >>> plt.ylabel('Gravitational potential')
        >>> plt.title('Gravitational Potential as a function of distance from point mass')
        >>> plt.show()
        """

        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)
        u_g = (G * self.mass) / r
        return u_g

    def gravitational_field(self, xyz):
        """
        Gravitational field due to a point mass.  See Blakely, 1996
        equation 3.3.

        .. math::

            \\mathbf{g} = \\nabla U(P)

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Point mass location in units m.

        Returns
        -------
        (..., 3) numpy.ndarray
            Gravitational field at point mass location xyz in units :math:`\\frac{m}{s^2}`.

        Examples
        --------
        Here, we define a point mass with mass=1kg and plot the gravitational
        field lines in the xy-plane.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from geoana.gravity import PointMass

        Define the point mass.

        >>> location = np.r_[0., 0., 0.]
        >>> mass = 1.0
        >>> simulation = PointMass(
        >>>     mass=mass, location=location
        >>> )

        Now we create a set of gridded locations and compute the gravitational field.

        >>> X, Y = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> g = simulation.gravitational_field(xyz)

        Finally, we plot the gravitational field lines.

        >>> plt.quiver(X, Y, g[:,:,0], g[:,:,1])
        >>> plt.xlabel('x')
        >>> plt.ylabel('y')
        >>> plt.title('Gravitational Field Lines for a Point Mass')
        >>> plt.show()
        """

        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)
        g_vec = -G * self.mass * r_vec / r[..., None] ** 3
        return g_vec

    def gravitational_gradient(self, xyz):
        """
        Gravitational gradient for a point mass.

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Point mass location in units m.

        Returns
        -------
        (..., 3, 3) numpy.ndarray
            Gravitational gradient at point mass location xyz in units :math:`\\frac{1}{s^2}`.

        Examples
        --------
        Here, we define a point mass with mass=1kg and plot the gravitational
        gradient.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from geoana.gravity import PointMass

        Define the point mass.

        >>> location = np.r_[0., 0., 0.]
        >>> mass = 1.0
        >>> simulation = PointMass(
        >>>     mass=mass, location=location
        >>> )

        Now we create a set of gridded locations and compute the gravitational gradient.

        >>> X, Y = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> g_tens = simulation.gravitational_gradient(xyz)

        Finally, we plot the gravitational gradient for each element of the 3 x 3 matrix.

        >>> fig = plt.figure(figsize=(10, 10))
        >>> gs = fig.add_gridspec(3, 3, hspace=0, wspace=0)
        >>> (ax1, ax2, ax3), (ax4, ax5, ax6), (ax7, ax8, ax9) = gs.subplots(sharex='col', sharey='row')
        >>> fig.suptitle('Gravitational Gradients for a Point Mass')
        >>> ax1.contourf(X, Y, g_tens[:,:,0,0])
        >>> ax2.contourf(X, Y, g_tens[:,:,0,1])
        >>> ax3.contourf(X, Y, g_tens[:,:,0,2])
        >>> ax4.contourf(X, Y, g_tens[:,:,1,0])
        >>> ax5.contourf(X, Y, g_tens[:,:,1,1])
        >>> ax6.contourf(X, Y, g_tens[:,:,1,2])
        >>> ax7.contourf(X, Y, g_tens[:,:,2,0])
        >>> ax8.contourf(X, Y, g_tens[:,:,2,1])
        >>> ax9.contourf(X, Y, g_tens[:,:,2,2])
        >>> plt.show()
        """

        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)
        g_tens = -G * self.mass * (np.eye(3) / r[..., None, None] ** 3 -
                                   3 * r_vec[..., None] * r_vec[..., None, :] / r[..., None, None] ** 5)
        return g_tens


class Sphere(PointMass):
    """Class for gravitational solutions for a sphere.

    The ``Sphere`` class is used to analytically compute the gravitational
    potentials, fields, and gradients for a sphere.

    Parameters
    ----------
    rho : float
        Density of sphere (:math:`\\frac{kg}{m^3}`).
    radius : float
        Radius of sphere (m).
    mass : float
        Mass of the sphere (kg). Default is :math:`m = \\frac{4}{3} \\pi R^3 \\rho` kg.
    location : array_like, optional
        Center of the sphere (m). Default is (0, 0, 0).
    """

    def __init__(self, radius, rho, location=None, **kwargs):
        self.radius = radius
        super().__init__(location=location, **kwargs)
        self.rho = rho

    @property
    def radius(self):
        """Radius of the sphere in meters.

        Returns
        -------
        float
            Radius of the sphere in meters.
        """
        return self._radius

    @radius.setter
    def radius(self, item):
        item = float(item)
        if item < 0.0:
            raise ValueError('radius must be non-negative')
        self._radius = item

    @property
    def rho(self):
        """Density of the sphere in kilograms over meters cubed.

        Returns
        -------
        float
            Density of the sphere in kilograms over meters cubed.
        """
        return self._rho

    @rho.setter
    def rho(self, item):
        item = float(item)
        self._rho = item

    @property
    def mass(self):
        """Mass of sphere in kilograms.

        Returns
        -------
        float
            Mass of the sphere in kilograms.
        """
        return 4 / 3 * np.pi * self.radius ** 3 * self.rho

    @mass.setter
    def mass(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"mass must be a number, got {type(value)}")

        rho = value * 3 / (4 * np.pi * self.radius ** 3)
        self._rho = rho

    def gravitational_potential(self, xyz):
        """
        Gravitational potential due to a sphere.

        .. math::

            r > R

            \\phi (\\mathbf{r}) = \\gamma \\frac{m}{r}

            r < R

            \\phi (\\mathbf{r}) = \\gamma \\frac{2}{3} \\pi \\rho (3R^2 - r^2)

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.

        Returns
        -------
        (..., ) numpy.ndarray
            Gravitational potential at sphere location xyz in units :math:`\\frac{m^2}{s^2}`.

        Examples
        --------
        Here, we define a sphere with mass m and plot the gravitational
        potential as a function of distance.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from geoana.gravity import Sphere
        >>> from geoana.utils import ndgrid

        Define the sphere.

        >>> location = np.r_[0., 0., 0.]
        >>> rho = 1.0
        >>> radius = 1.0
        >>> simulation = Sphere(
        >>>     location=location, rho=rho, radius=radius
        >>> )

        Now we create a set of gridded locations, take the distances and compute the gravitational potential.

        >>> X, Y = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> r = np.linalg.norm(xyz, axis=-1)
        >>> u = simulation.gravitational_potential(xyz)

        Finally, we plot the gravitational potential as a function of distance.

        >>> plt.plot(r, u)
        >>> plt.xlabel('Distance from sphere')
        >>> plt.ylabel('Gravitational potential')
        >>> plt.title('Gravitational Potential as a function of distance from sphere')
        >>> plt.show()
        """

        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)
        u_g = np.zeros_like(r)
        ind0 = r > self.radius
        u_g[ind0] = super().gravitational_potential(xyz[ind0])
        u_g[~ind0] = G * 2 / 3 * np.pi * self.rho * (3 * self.radius ** 2 - r[~ind0] ** 2)
        return u_g

    def gravitational_field(self, xyz):
        """
        Gravitational field due to a sphere.

        .. math::

            r > R

            \\mathbf{g} (\\mathbf{r}) = \\nabla \\phi (\\mathbf{r})

            r < R

            \\mathbf{g} (\\mathbf{r}) = - \\gamma \\frac{4}{3} \\pi \\rho \\mathbf{r}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.

        Returns
        -------
        (..., 3) numpy.ndarray
            Gravitational field at sphere location xyz in units :math:`\\frac{m}{s^2}`.

        Examples
        --------
        Here, we define a sphere with mass m and plot the gravitational
        field lines in the xy-plane.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from geoana.gravity import Sphere

        Define the sphere.

        >>> location = np.r_[0., 0., 0.]
        >>> rho = 1.0
        >>> radius = 1.0
        >>> simulation = Sphere(
        >>>     location=location, rho=rho, radius=radius
        >>> )

        Now we create a set of gridded locations and compute the gravitational field.

        >>> X, Y = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> g = simulation.gravitational_field(xyz)

        Finally, we plot the gravitational field lines.

        >>> plt.quiver(X, Y, g[:,:,0], g[:,:,1])
        >>> plt.xlabel('x')
        >>> plt.ylabel('y')
        >>> plt.title('Gravitational Field Lines for a Sphere')
        >>> plt.show()
        """

        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)
        g_vec = np.zeros((*r.shape, 3))
        ind0 = r > self.radius
        g_vec[ind0] = super().gravitational_field(xyz[ind0])
        g_vec[~ind0] = -G * 4 / 3 * np.pi * self.rho * r_vec[~ind0]
        return g_vec

    def gravitational_gradient(self, xyz):
        """
        Gravitational gradient for a sphere.

        .. math::

            r < R

            \\mathbf{T} (\\mathbf{r}) = -\\gamma \\frac{4}{3} \\pi \\rho \\mathbf{I}

        Parameters
        ----------
        xyz : (..., 3) numpy.ndarray
            Locations to evaluate at in units m.

        Returns
        -------
        (..., 3, 3) numpy.ndarray
            Gravitational gradient at sphere location xyz in units :math:`\\frac{1}{s^2}`.

        Examples
        --------
        Here, we define a sphere with mass m and plot the gravitational
        gradient.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from geoana.gravity import Sphere

        Define the sphere.

        >>> location = np.r_[0., 0., 0.]
        >>> rho = 1.0
        >>> radius = 1.0
        >>> simulation = Sphere(
        >>>     location=location, rho=rho, radius=radius
        >>> )

        Now we create a set of gridded locations and compute the gravitational gradient.

        >>> X, Y = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))
        >>> Z = np.zeros_like(X) + 0.25
        >>> xyz = np.stack((X, Y, Z), axis=-1)
        >>> g_tens = simulation.gravitational_gradient(xyz)

        Finally, we plot the gravitational gradient for each element of the 3 x 3 matrix.

        >>> fig = plt.figure(figsize=(10, 10))
        >>> gs = fig.add_gridspec(3, 3, hspace=0, wspace=0)
        >>> (ax1, ax2, ax3), (ax4, ax5, ax6), (ax7, ax8, ax9) = gs.subplots(sharex='col', sharey='row')
        >>> fig.suptitle('Gravitational Gradients for a Sphere')
        >>> ax1.contourf(X, Y, g_tens[:,:,0,0])
        >>> ax2.contourf(X, Y, g_tens[:,:,0,1])
        >>> ax3.contourf(X, Y, g_tens[:,:,0,2])
        >>> ax4.contourf(X, Y, g_tens[:,:,1,0])
        >>> ax5.contourf(X, Y, g_tens[:,:,1,1])
        >>> ax6.contourf(X, Y, g_tens[:,:,1,2])
        >>> ax7.contourf(X, Y, g_tens[:,:,2,0])
        >>> ax8.contourf(X, Y, g_tens[:,:,2,1])
        >>> ax9.contourf(X, Y, g_tens[:,:,2,2])
        >>> plt.show()
        """

        r_vec = xyz - self.location
        r = np.linalg.norm(r_vec, axis=-1)
        g_tens = np.zeros((*r.shape, 3, 3))
        ind0 = r > self.radius
        ind1 = r == self.radius
        g_tens[ind0] = super().gravitational_gradient(xyz[ind0])
        g_tens[~ind0] = -G * 4 / 3 * np.pi * self.rho * np.eye(3)
        g_tens[ind1] = np.NaN
        return g_tens



"""
======================================================
Time-Domain Electromagnetics (:mod:`geoana.em.tdem`)
======================================================
.. currentmodule:: geoana.em.tdem

The ``geoana.em.tdem`` module contains simulation classes for solving
basic time-domain electromagnetic problems.

Simulation Classes
==================
.. autosummary::
  :toctree: generated/

  BaseTDEM
  ElectricDipoleWholeSpace
  VerticalMagneticDipoleHalfSpace

Utility Functions
=================

.. autosummary::
  :toctree: generated/

  peak_time
  diffusion_distance
  theta
  vertical_magnetic_field_horizontal_loop
  vertical_magnetic_flux_horizontal_loop
  vertical_magnetic_field_time_deriv_horizontal_loop
  vertical_magnetic_flux_time_deriv_horizontal_loop
  magnetic_field_vertical_magnetic_dipole
  magnetic_field_time_deriv_magnetic_dipole
  magnetic_flux_vertical_magnetic_dipole
  magnetic_flux_time_deriv_magnetic_dipole
  
"""
from geoana.em.tdem.base import peak_time, diffusion_distance, theta, BaseTDEM

from geoana.em.tdem.wholespace import ElectricDipoleWholeSpace

from geoana.em.tdem.halfspace import VerticalMagneticDipoleHalfSpace

from geoana.em.tdem.simple_functions import (
    vertical_magnetic_field_horizontal_loop,
    vertical_magnetic_flux_horizontal_loop,
    vertical_magnetic_field_time_deriv_horizontal_loop,
    vertical_magnetic_flux_time_deriv_horizontal_loop,
    magnetic_field_vertical_magnetic_dipole,
    magnetic_field_time_deriv_magnetic_dipole,
    magnetic_flux_vertical_magnetic_dipole,
    magnetic_flux_time_deriv_magnetic_dipole,
)



"""
=================================================
Spatial Utilities (:mod:`geoana.spatial`)
=================================================
.. currentmodule:: geoana.spatial

The ``geoana.spatial`` module provides rudimentary functions for vectors and
functions for converting between Cartesian, cylindrical and spherical coordinates.


.. autosummary::
  :toctree: generated/

  cylindrical_to_cartesian
  cartesian_to_cylindrical
  vector_magnitude
  vector_distance
  distance
  vector_dot
  repeat_scalar
  rotation_matrix_from_normals
  rotate_points_from_normals

"""
import numpy as np
from .utils import mkvc


def cylindrical_to_cartesian(grid, vec=None):
    """
    Transform gridded locations or a set of vectors from cylindrical coordinates
    :math:`(r, \\phi, z)` to Cartesian coordinates :math:`(x, y, z)`.
    The azimuthal angle :math:`\\phi` is given in radians.

    Parameters
    ----------
    grid : (n, 3) array_like
        Location points defined in cylindrical coordinates :math:`(r, \\phi, z)`.
    vec : (n, 3) array_like, optional
        Vectors defined in cylindrical coordinates :math:`(v_r, v_\\phi, v_z)` at the
        gridded locations. Will also except a flattend array in column major order
        with the same number of elements.

    Returns
    -------
    (n, 3) numpy.ndarray
        If `vec` is ``None``, this returns the transformed `grid` array, otherwise
        this is the transformed `vec` array.

    Examples
    --------
    Here, we convert a series of vectors in 3D space from cylindrical coordinates
    to Cartesian coordinates.

    >>> from geoana.spatial import cylindrical_to_cartesian
    >>> import numpy as np

    Construct original set of vectors in cylindrical coordinates

    >>> r = np.ones(9)
    >>> phi = np.linspace(0, 2*np.pi, 9)
    >>> z = np.linspace(-4., 4., 9)
    >>> u = np.c_[r, phi, z]
    >>> u
    array([[ 1.        ,  0.        , -4.        ],
           [ 1.        ,  0.78539816, -3.        ],
           [ 1.        ,  1.57079633, -2.        ],
           [ 1.        ,  2.35619449, -1.        ],
           [ 1.        ,  3.14159265,  0.        ],
           [ 1.        ,  3.92699082,  1.        ],
           [ 1.        ,  4.71238898,  2.        ],
           [ 1.        ,  5.49778714,  3.        ],
           [ 1.        ,  6.28318531,  4.        ]])

    Create equivalent set of vectors in Cartesian coordinates

    >>> v = cylindrical_to_cartesian(u)
    >>> v
    array([[ 1.00000000e+00,  0.00000000e+00, -4.00000000e+00],
           [ 7.07106781e-01,  7.07106781e-01, -3.00000000e+00],
           [ 6.12323400e-17,  1.00000000e+00, -2.00000000e+00],
           [-7.07106781e-01,  7.07106781e-01, -1.00000000e+00],
           [-1.00000000e+00,  1.22464680e-16,  0.00000000e+00],
           [-7.07106781e-01, -7.07106781e-01,  1.00000000e+00],
           [-1.83697020e-16, -1.00000000e+00,  2.00000000e+00],
           [ 7.07106781e-01, -7.07106781e-01,  3.00000000e+00],
           [ 1.00000000e+00, -2.44929360e-16,  4.00000000e+00]])
    """
    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(grid[:, 0]*np.cos(grid[:, 1]), 2),
            mkvc(grid[:, 0]*np.sin(grid[:, 1]), 2),
            mkvc(grid[:, 2], 2)
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    x = vec[:, 0] * np.cos(grid[:, 1]) - vec[:, 1] * np.sin(grid[:, 1])
    y = vec[:, 0] * np.sin(grid[:, 1]) + vec[:, 1] * np.cos(grid[:, 1])

    newvec = [x, y]
    if grid.shape[1] == 3:
        z = vec[:, 2]
        newvec += [z]

    return np.vstack(newvec).T


def cartesian_to_cylindrical(grid, vec=None):
    """
    Transform gridded locations or a set of vectors from Cartesian coordinates
    :math:`(x, y, z)` to cylindrical coordinates :math:`(r, \\phi, z)`. Where
    the azimuthal angle :math:`\\phi \\in [-\\pi , \\pi ]` will be given output
    in radians.

    Parameters
    ----------
    grid : (n, 3) array_like
        Gridded locations defined in Cartesian coordinates :math:`(x, y z)`.
    vec : (n, 3) array_like, optional
        Vectors defined in Cartesian coordinates :math:`(v_x, v_y, v_z)` at the
        gridded locations. Also accepts a flattened array with the same total
        elements in column major order.

    Returns
    -------
    (n, 3) numpy.ndarray
        If `vec` is ``None``, this returns the transformed `grid` array, otherwise
        this is the transformed `vec` array.

    Examples
    --------
    Here, we convert a series of vectors in 3D space from Cartesian coordinates
    to cylindrical coordinates.

    >>> from geoana.spatial import cartesian_to_cylindrical
    >>> import numpy as np

    Create set of vectors in Cartesian coordinates

    >>> r = np.ones(9)
    >>> phi = np.linspace(0, 2*np.pi, 9)
    >>> z = np.linspace(-4., 4., 9)
    >>> x = r*np.cos(phi)
    >>> y = r*np.sin(phi)
    >>> u = np.c_[x, y, z]
    >>> u
    array([[ 1.00000000e+00,  0.00000000e+00, -4.00000000e+00],
           [ 7.07106781e-01,  7.07106781e-01, -3.00000000e+00],
           [ 6.12323400e-17,  1.00000000e+00, -2.00000000e+00],
           [-7.07106781e-01,  7.07106781e-01, -1.00000000e+00],
           [-1.00000000e+00,  1.22464680e-16,  0.00000000e+00],
           [-7.07106781e-01, -7.07106781e-01,  1.00000000e+00],
           [-1.83697020e-16, -1.00000000e+00,  2.00000000e+00],
           [ 7.07106781e-01, -7.07106781e-01,  3.00000000e+00],
           [ 1.00000000e+00, -2.44929360e-16,  4.00000000e+00]])

    Compute equivalent set of vectors in cylindrical coordinates

    >>> v = cartesian_to_cylindrical(u)
    >>> v
    array([[ 1.00000000e+00,  0.00000000e+00, -4.00000000e+00],
           [ 1.00000000e+00,  7.85398163e-01, -3.00000000e+00],
           [ 1.00000000e+00,  1.57079633e+00, -2.00000000e+00],
           [ 1.00000000e+00,  2.35619449e+00, -1.00000000e+00],
           [ 1.00000000e+00,  3.14159265e+00,  0.00000000e+00],
           [ 1.00000000e+00, -2.35619449e+00,  1.00000000e+00],
           [ 1.00000000e+00, -1.57079633e+00,  2.00000000e+00],
           [ 1.00000000e+00, -7.85398163e-01,  3.00000000e+00],
           [ 1.00000000e+00, -2.44929360e-16,  4.00000000e+00]])
    """

    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2), 2),
            mkvc(np.arctan2(grid[:, 1], grid[:, 0]), 2),
            mkvc(grid[:, 2], 2)
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    theta = np.arctan2(grid[:, 1], grid[:, 0])

    return np.hstack([
        mkvc(np.cos(theta)*vec[:, 0] + np.sin(theta)*vec[:, 1], 2),
        mkvc(-np.sin(theta)*vec[:, 0] + np.cos(theta)*vec[:, 1], 2),
        mkvc(vec[:, 2], 2)
    ])


def spherical_to_cartesian(grid, vec=None):
    """
    Transform gridded locations of a set of vectors from spherical coordinates
    :math:`(r, \\phi, \\theta)` to Cartesian coordinates :math:`(x, y, z)`.
    :math:`\\phi` and :math:`\\theta` are the azimuthal and polar angles, respectively.
    :math:`\\phi` and :math:`\\theta` are given in radians.

    Parameters
    ----------
    grid : (n, 3) array_like
        Gridded locations defined in spherical coordinates :math:`(r, \\phi, \\theta)`.
    vec : (n, 3) array_like, optional
        Vectors defined in spherical coordinates :math:`(v_r, v_\\phi, v_\\theta)` at the
        gridded locations. Will also except a flattend array in column major order with the
        same number of elements.

    Returns
    -------
    (n, 3) numpy.ndarray
        If `vec` is ``None``, this returns the transformed `grid` array, otherwise
        this is the transformed `vec` array.

    Examples
    --------
    Here, we convert a series of vectors in 3D space from spherical coordinates
    to Cartesian coordinates.

    >>> from geoana.spatial import spherical_to_cartesian
    >>> import numpy as np

    Construct original set of vectors in spherical coordinates

    >>> r = np.ones(9)
    >>> phi = np.linspace(0, 2*np.pi, 9)
    >>> theta = np.linspace(0, np.pi, 9)
    >>> u = np.c_[r, phi, theta]
    >>> u
    array([[1.        , 0.        , 0.        ],
           [1.        , 0.78539816, 0.39269908],
           [1.        , 1.57079633, 0.78539816],
           [1.        , 2.35619449, 1.17809725],
           [1.        , 3.14159265, 1.57079633],
           [1.        , 3.92699082, 1.96349541],
           [1.        , 4.71238898, 2.35619449],
           [1.        , 5.49778714, 2.74889357],
           [1.        , 6.28318531, 3.14159265]])

    Create equivalent set of vectors in Cartesian coordinates

    >>> v = spherical_to_cartesian(u)
    >>> v
    array([[ 0.00000000e+00,  0.00000000e+00,  1.00000000e+00],
           [ 2.70598050e-01,  2.70598050e-01,  9.23879533e-01],
           [ 4.32978028e-17,  7.07106781e-01,  7.07106781e-01],
           [-6.53281482e-01,  6.53281482e-01,  3.82683432e-01],
           [-1.00000000e+00,  1.22464680e-16,  6.12323400e-17],
           [-6.53281482e-01, -6.53281482e-01, -3.82683432e-01],
           [-1.29893408e-16, -7.07106781e-01, -7.07106781e-01],
           [ 2.70598050e-01, -2.70598050e-01, -9.23879533e-01],
           [ 1.22464680e-16, -2.99951957e-32, -1.00000000e+00]])
    """
    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(grid[:, 0] * np.sin(grid[:, 2]) * np.cos(grid[:, 1]), 2),
            mkvc(grid[:, 0] * np.sin(grid[:, 2]) * np.sin(grid[:, 1]), 2),
            mkvc(grid[:, 0] * np.cos(grid[:, 2]), 2)
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    x = (
        vec[:, 0] * np.sin(grid[:, 2]) * np.cos(grid[:, 1]) +
        vec[:, 2] * np.cos(grid[:, 2]) * np.cos(grid[:, 1]) -
        vec[:, 1] * np.sin(grid[:, 1])
    )
    y = (
        vec[:, 0] * np.sin(grid[:, 2]) * np.sin(grid[:, 1]) +
        vec[:, 2] * np.cos(grid[:, 2]) * np.sin(grid[:, 1]) -
        vec[:, 1] * np.cos(grid[:, 1])
    )
    z = (
        vec[:, 0] * np.cos(grid[:, 2]) -
        vec[:, 2] * np.sin(grid[:, 2])
    )

    newvec = [x, y, z]

    return np.vstack(newvec).T


def cartesian_to_spherical(grid, vec=None):
    """
    Transform gridded locations or a set of vectors from Cartesian coordinates
    :math:`(x, y, z)` to  spherical coordinates :math:`(r, \\phi, \\theta)`.
    :math:`\\phi` and :math:`\\theta` are the azimuthal and polar angle, respectively.
    :math:`\\phi` and :math:`\\theta` are given in radians.

    Parameters
    ----------
    grid : (n, 3) array_like
        Gridded locations defined in Cartesian coordinates :math:`(x, y, z)`.
    vec : (n, 3) array_like, optional
        Vectors defined in Cartesian coordinates :math:`(v_x, v_y, v_z)` at the
        gridded locations. Will also except a flattend array in column major order with the
        same number of elements.

    Returns
    -------
    (n, 3) numpy.ndarray
        If `vec` is ``None``, this returns the transformed `grid` array, otherwise
        this is the transformed `vec` array.

    Examples
    --------
    Here, we convert a series of vectors in 3D space from Cartesian coordinates
    to spherical coordinates.

    >>> from geoana.spatial import cartesian_to_spherical
    >>> import numpy as np

    Construct original set of vectors in cartesian coordinates

    >>> r = np.ones(9)
    >>> phi = np.linspace(0, 2*np.pi, 9)
    >>> theta = np.linspace(0., np.pi, 9)
    >>> x = r*np.sin(theta)*np.cos(phi)
    >>> y = r*np.sin(theta)*np.sin(phi)
    >>> z = r*np.cos(theta)
    >>> u = np.c_[x, y, z]
    >>> u
    array([[ 0.00000000e+00,  0.00000000e+00,  1.00000000e+00],
           [ 2.70598050e-01,  2.70598050e-01,  9.23879533e-01],
           [ 4.32978028e-17,  7.07106781e-01,  7.07106781e-01],
           [-6.53281482e-01,  6.53281482e-01,  3.82683432e-01],
           [-1.00000000e+00,  1.22464680e-16,  6.12323400e-17],
           [-6.53281482e-01, -6.53281482e-01, -3.82683432e-01],
           [-1.29893408e-16, -7.07106781e-01, -7.07106781e-01],
           [ 2.70598050e-01, -2.70598050e-01, -9.23879533e-01],
           [ 1.22464680e-16, -2.99951957e-32, -1.00000000e+00]])

    Compute equivalent set of vectors in spherical coordinates

    >>> v = cartesian_to_spherical(u)
    >>> v
    array([[ 1.00000000e+00,  0.00000000e+00,  0.00000000e+00],
           [ 1.00000000e+00,  7.85398163e-01,  3.92699082e-01],
           [ 1.00000000e+00,  1.57079633e+00,  7.85398163e-01],
           [ 1.00000000e+00,  2.35619449e+00,  1.17809725e+00],
           [ 1.00000000e+00,  3.14159265e+00,  1.57079633e+00],
           [ 1.00000000e+00, -2.35619449e+00,  1.96349541e+00],
           [ 1.00000000e+00, -1.57079633e+00,  2.35619449e+00],
           [ 1.00000000e+00, -7.85398163e-01,  2.74889357e+00],
           [ 1.00000000e+00, -2.44929360e-16,  3.14159265e+00]])
    """

    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2 + grid[:, 2]**2), 2),
            mkvc(np.arctan2(grid[:, 1], grid[:, 0]), 2),
            mkvc(
                np.arctan2(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2), grid[:, 2]),
                2
            ),
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    theta = np.arctan2(grid[:, 1], grid[:, 0])
    phi = np.arctan2(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2), grid[:, 2])

    r = (
        vec[:, 0] * np.sin(phi) * np.cos(theta) +
        vec[:, 1] * np.sin(phi) * np.sin(theta) +
        vec[:, 2] * np.cos(phi)
    )

    theta = - vec[:, 0] * np.sin(theta) + vec[:, 1] * np.cos(theta)

    phi = (
        vec[:, 0] * np.cos(phi) * np.cos(theta) +
        vec[:, 1] * np.cos(phi) * np.sin(theta) -
        vec[:, 2] * np.sin(phi)
    )

    newvec = [r, theta, phi]

    return np.vstack(newvec).T


def vector_magnitude(v):
    """Compute the amplitudes for a set of input vectors in Cartesian coordinates.

    Parameters
    ----------
    (n, dim) numpy.ndarray
        A set of input vectors (2D or 3D)

    Returns
    -------
    (n) numpy.ndarray
        Magnitudes of the vectors
    """

    v = np.atleast_2d(v)

    return np.sqrt((v**2).sum(axis=1))


def vector_distance(xyz, origin=np.r_[0., 0., 0.]):
    r"""Vector distances from a reference location to a set of xyz locations.

    Where :math:`\mathbf{p}` is an reference location define by the input argument
    `origin`, this method returns the vector distance:

    .. math::
        \mathbf{v} = \mathbf{q} - \mathbf{p}

    for all locations :math:`\mathbf{q}` supplied in the input argument `xyz`.
    By default, the reference location is (0,0,0).

    Parameters
    ----------
    xyz : (n, 3) numpy.ndarray
        Gridded xyz locations
    origin : (3) array_like, optional
        Reference location. Default = np.r_[0,0,0]

    Returns
    -------
    (n, 3) numpy.ndarray
        Vector distances along the x, y and z directions

    """
    assert(xyz.shape[1] == 3), (
        "the xyz grid should be npoints by 3, the shape provided is {}".format(
            xyz.shape
        )
    )

    if len(origin) != 3:
        raise Exception(
            "the origin must be length 3, the length provided is {}".format(
                len(origin)
            )
        )


    dx = xyz[:, 0] - origin[0]
    dy = xyz[:, 1] - origin[1]
    dz = xyz[:, 2] - origin[2]

    return np.c_[dx, dy, dz]


def distance(xyz, origin=np.r_[0., 0., 0.]):
    r"""Scalar distances between a reference location to a set of xyz locations.

    Where :math:`\mathbf{p}` is an reference location define by the input argument
    `origin`, this method returns the scalar distance:

    .. math::
        d = \big | \mathbf{q} - \mathbf{p} \big |

    for all locations :math:`\mathbf{q}` supplied in the input argument `xyz`.
    By default, the reference location is (0,0,0).

    Parameters
    ----------
    xyz : (n, 3) numpy.ndarray
        Gridded xyz locations
    origin : (3) array_like (optional)
        Reference location. Default = np.r_[0,0,0]

    Returns
    -------
    (n) numpy.ndarray
        Scalar distances
    """
    dxyz = vector_distance(xyz, origin)
    return vector_magnitude(dxyz)


def vector_dot(xyz, vector):
    r"""Dot product between a vector and a set of vectors.

    Where :math:`\mathbf{u}` is a vector defined by the input argument `vector`,
    this method returns the dot products between :math:`\mathbf{u}` and every
    vector supplied in the input argument `xyz`. I.e. for all
    :math:`\mathbf{v} \\in \mathbf{V}`, we compute:

    .. math::
        \mathbf{u} \cdot \mathbf{v}

    Parameters
    ----------
    xyz : (n, 3) numpy.ndarray
        A set of 3D vectors
    vector : (3) numpy.array_like
        A single 3D vector

    Returns
    -------
    (n) numpy.ndarray
        Dot products between a vector and a set of vectors.
    """
    if len(vector) != 3:
        raise Exception(
            "vector should be length 3, the provided length is {}".format(
                len(vector)
            )
        )
    return vector[0]*xyz[:, 0] + vector[1]*xyz[:, 1] + vector[2]*xyz[:, 2]


def repeat_scalar(scalar, dim=3):
    """Stack spatially distributed scalar values to simplify multiplication with a vector.

    Where the input argument `scalar` defines a set of *n* spatially distributed scalar
    values, ``repeat_scalar`` repeats and stacks the input array to produce an
    (`n`, `dim`) array which is better for multiplying the scalar values with vector arrays.

    Parameters
    ----------
    scalar : (n) numpy.ndarray
        The set of scalar values
    dim : int, optional
        The dimension. Default=3

    Returns
    -------
    (n, dim) numpy.ndarray
        The repeated and stacked set of scalar values
    """
    assert len(scalar) in scalar.shape, (
        "input must be a scalar. The shape you provided is {}".format(
            scalar.shape
        )
    )

    return np.kron(np.ones((1, dim)), np.atleast_2d(scalar).T)

def rotation_matrix_from_normals(v0, v1, tol=1e-20):
    """
    Generate a 3x3 rotation matrix defining the rotation from vector v0 to v1.

    This function uses Rodrigues' rotation formula to generate the rotation
    matrix :math:`\\mathbf{A}` going from vector :math:`\\mathbf{v_0}` to
    vector :math:`\\mathbf{v_1}`. Thus:

    .. math::
        \\mathbf{Av_0} = \\mathbf{v_1}

    For detailed desciption of the algorithm, see
    https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula

    Parameters
    ----------
    v0 : (3) numpy.ndarray
        Starting orientation direction
    v1 : (3) numpy.ndarray
        Finishing orientation direction
    tol : float, optional
        Numerical tolerance. If the length of the rotation axis is below this value,
        it is assumed to be no rotation, and an identity matrix is returned.

    Returns
    -------
    (3, 3) numpy.ndarray
        The rotation matrix from v0 to v1.
    """

    v0 = mkvc(v0)
    v1 = mkvc(v1)

    # ensure both n0, n1 are vectors of length 1
    assert len(v0) == 3, "Length of n0 should be 3"
    assert len(v1) == 3, "Length of n1 should be 3"

    # ensure both are true normals
    n0 = v0*1./np.linalg.norm(v0)
    n1 = v1*1./np.linalg.norm(v1)

    n0dotn1 = n0.dot(n1)

    # define the rotation axis, which is the cross product of the two vectors
    rotAx = np.cross(n0, n1)

    if np.linalg.norm(rotAx) < tol:
        return np.eye(3, dtype=float)

    rotAx *= 1./np.linalg.norm(rotAx)

    cosT = n0dotn1/(np.linalg.norm(n0)*np.linalg.norm(n1))
    sinT = np.sqrt(1.-n0dotn1**2)

    ux = np.array(
        [
            [0., -rotAx[2], rotAx[1]],
            [rotAx[2], 0., -rotAx[0]],
            [-rotAx[1], rotAx[0], 0.]
        ], dtype=float
    )

    return np.eye(3, dtype=float) + sinT*ux + (1.-cosT)*(ux.dot(ux))


def rotate_points_from_normals(xyz, n0, n1, x0=np.r_[0., 0., 0.]):
    """Rotate a set of xyz locations about a specified point.

    Rotate a grid of Cartesian points about a location x0 according to the
    rotation defined from vector v0 to v1.

    Let :math:`\\mathbf{x}` represent an input xyz location, let :math:`\\mathbf{x_0}` be
    the origin of rotation, and let :math:`\\mathbf{R}` denote the rotation matrix from
    vector v0 to v1. Where :math:`\\mathbf{x'}` is the new xyz location, this function
    outputs the following operation for all input locations:

    .. math::
        \\mathbf{x'} = \\mathbf{R (x - x_0)} + \\mathbf{x_0}

    Parameters
    ----------
    xyz : (n, 3) numpy.ndarray
        locations to rotate
    v0 : (3) numpy.ndarray
        Starting orientation direction
    v1 : (3) numpy.ndarray
        Finishing orientation direction
    x0 : (3) numpy.ndarray, optional
        The origin of rotation.

    Returns
    -------
    (n, 3) numpy.ndarray
        The rotated xyz locations.
    """

    R = rotation_matrix_from_normals(n0, n1)

    if xyz.shape[1] != 3:
        raise AssertionError("Grid xyz should be 3 wide")

    return (xyz - x0)@R.T + x0



# Aliases
def cylindrical_2_cartesian(grid, vec=None):
    """An alias for cylindrical_to_cartesian"""
    return cylindrical_to_cartesian(grid, vec)

def cartesian_2_cylindrical(grid, vec=None):
    """An alias for cartesian_to_cylindrical"""
    return cartesian_to_cylindrical(grid, vec)

def spherical_2_cartesian(grid, vec=None):
    """An alias for spherical_to_cartesian"""
    return spherical_to_cartesian(grid, vec)

def cartesian_2_spherical(grid, vec=None):
    """An Alias for cartesian_to_spherical"""
    return cartesian_to_spherical(grid, vec)



"""
============================================================
Frequency-Domain Electromagnetics (:mod:`geoana.em.fdem`)
============================================================
.. currentmodule:: geoana.em.fdem

The ``geoana.em.fdem`` module contains simulation classes for solving
basic frequency-domain electromagnetic problems.

Simulation Classes
==================
.. autosummary::
  :toctree: generated/

  BaseFDEM
  ElectricDipoleWholeSpace
  MagneticDipoleWholeSpace
  MagneticDipoleHalfSpace
  MagneticDipoleLayeredHalfSpace

Utility Functions
=================

.. autosummary::
  :toctree: generated/

  omega
  wavenumber
  skin_depth
  sigma_hat
  vertical_magnetic_field_horizontal_loop
  vertical_magnetic_flux_horizontal_loop
"""
from geoana.em.fdem.base import (
    omega, wavenumber, skin_depth, sigma_hat, BaseFDEM
)

from geoana.em.fdem.wholespace import (
    ElectricDipoleWholeSpace, MagneticDipoleWholeSpace
)

from geoana.em.fdem.halfspace import MagneticDipoleHalfSpace

from geoana.em.fdem.layered import MagneticDipoleLayeredHalfSpace

from geoana.em.fdem.simple_functions import (
    vertical_magnetic_field_horizontal_loop, vertical_magnetic_flux_horizontal_loop
)



from scipy.special import erf
import numpy as np
from geoana.em.tdem.base import BaseTDEM
from geoana.spatial import repeat_scalar
from geoana.em.base import BaseElectricDipole

###############################################################################
#                                                                             #
#                                  Classes                                    #
#                                                                             #
###############################################################################


class ElectricDipoleWholeSpace(BaseTDEM, BaseElectricDipole):
    r"""
    Class for simulating the fields and fluxes for a transient electric
    dipole in a wholespace.

    """

    def vector_potential(self, xyz):
        r"""Vector potential for the transient current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}`
        direction with dipole moment :math:`I ds`, this method computes
        the transient vector potential at the set of observation times for
        the gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988).
        For a transient electric current dipole oriented in the
        :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \mathbf{a}(t) = \frac{I ds}{4 \pi r} \, erf(\theta r) \, \hat{x}

        where

        .. math::
            \theta = \Bigg ( \frac{\mu\sigma}{4t} \Bigg )^{1/2}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_time, n_loc, 3) numpy.array of float
            Vector potential at all times for the gridded
            locations provided. Output array is squeezed when n_time and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an x-oriented electric dipole and plot the vector
        potential on the xz-plane that intercepts y=0.

        >>> from geoana.em.tdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> time = np.logspace(-6, -2, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[1., 0., 0.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     time, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the vector potential.

        >>> xyz = ndgrid(np.linspace(-10, 10, 20), np.array([0]), np.linspace(-10, 10, 20))
        >>> E = simulation.vector_potential(xyz)

        Finally, we plot the vector potential at the desired locations/times.

        >>> t_ind = 0
        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0::2], E[t_ind, :, 0::2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Vector potential at {} s'.format(time[t_ind]))
        """

        r = self.distance(xyz)

        n_loc = len(r)
        n_time = len(self.time)

        theta_r = np.outer(self.theta, r)
        tile_r = np.outer(np.ones(n_time), r)

        term_1 = (
            (self.current * self.length) * erf(theta_r) / (4 * np.pi * tile_r**3)
        ).reshape((n_time, n_loc, 1))
        term_1 = np.tile(term_1, (1, 1, 3))

        term_2 = np.tile(np.reshape(self.orientation, (1, 1, 3)), (n_time, n_loc, 1))

        return (term_1 * term_2).squeeze()

    def electric_field(self, xyz):
        r"""Electric field for the transient current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}`
        direction with dipole moment :math:`I ds`, this method computes
        the transient electric field at the set of observation times for
        the gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988).
        For a transient electric current dipole oriented in the
        :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            {\bf e_e}(t) = \frac{Ids}{4\pi \sigma r^3} & \Bigg [ \Bigg ( \frac{x^2}{r^2}\mathbf{\hat x} + \frac{xy}{r^2}\mathbf{\hat y} + \frac{xz}{r^2}\mathbf{\hat z} \Bigg ) ... \\
            & \Bigg ( 3 \, \textrm{erf}(\theta r) - \bigg ( \frac{4}{\sqrt{\pi}}\theta^3 r^3 + \frac{6}{\sqrt{\pi}} \theta r \bigg ) e^{-\theta^2 r^2}  \Bigg ) - \Bigg ( \textrm{erf}(\theta r) - \bigg ( \frac{4}{\sqrt{\pi}} \theta^3 r^3 + \frac{2}{\sqrt{\pi}} \theta r \bigg ) e^{-\theta^2 r^2} \Bigg ) \mathbf{\hat x} \Bigg ]

        where

        .. math::
            \theta = \Bigg ( \frac{\mu\sigma}{4t} \Bigg )^{1/2}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_time, n_loc, 3) numpy.array of float
            Electric field at all times for the gridded
            locations provided. Output array is squeezed when n_time and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an x-oriented electric dipole and plot the electric
        field on the xz-plane that intercepts y=0.

        >>> from geoana.em.tdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> time = np.logspace(-6, -2, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[1., 0., 0.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     time, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the electric field.

        >>> xyz = ndgrid(np.linspace(-10, 10, 20), np.array([0]), np.linspace(-10, 10, 20))
        >>> E = simulation.electric_field(xyz)

        Finally, we plot the electric field at the desired locations/times.

        >>> t_ind = 0
        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0::2], E[t_ind, :, 0::2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Electric field at {} s'.format(time[t_ind]))

        """
        # dxyz = self.vector_distance(xyz)
        # r = self.distance(xyz)
        # r = repeat_scalar(r)
        # theta_r = self.theta * r
        # root_pi = np.sqrt(np.pi)

        # front = (
        #     (self.current * self.length) / (4 * np.pi * self.sigma * r**3)
        # )

        # symmetric_term = (
        #     (
        #         - (
        #             4/root_pi * theta_r ** 3 + 6/root_pi * theta_r
        #         ) * np.exp(-theta_r**2) +
        #         3 * erf(theta_r)
        #     ) * (
        #         repeat_scalar(self.dot_orientation(dxyz)) * dxyz / r**2
        #     )
        # )

        # oriented_term = (
        #     (
        #         4./root_pi * theta_r**3 + 2./root_pi * theta_r
        #     ) * np.exp(-theta_r**2) -
        #     erf(theta_r)
        # ) * np.kron(self.orientation, np.ones((dxyz.shape[0], 1)))

        # return front * (symmetric_term + oriented_term)

        root_pi = np.sqrt(np.pi)
        dxyz = self.vector_distance(xyz)
        r = self.distance(xyz)

        n_loc = len(r)
        n_time = len(self.time)

        theta_r = np.outer(self.theta, r)
        tile_r = np.outer(np.ones(n_time), r)
        r = repeat_scalar(r)

        front = (
            (self.current * self.length) / (4 * np.pi * self.sigma * tile_r**3)
        ).reshape((n_time, n_loc, 1))
        front = np.tile(front, (1, 1, 3))

        term_1 = 3 * erf(theta_r) - (4/root_pi * theta_r ** 3 + 6/root_pi * theta_r) * np.exp(-theta_r**2)
        term_1 = np.tile(term_1.reshape((n_time, n_loc, 1)), (1, 1, 3))
        term_2 = repeat_scalar(self.dot_orientation(dxyz)) * dxyz / r**2
        term_2 = np.tile(term_2.reshape((1, n_loc, 3)), (n_time, 1, 1))
        symmetric_term = term_1 * term_2

        term_1 = (4./root_pi * theta_r**3 + 2./root_pi * theta_r) * np.exp(-theta_r**2) - erf(theta_r)
        term_1 = np.tile(term_1.reshape((n_time, n_loc, 1)), (1, 1, 3))
        term_2 = np.kron(self.orientation, np.ones((dxyz.shape[0], 1)))
        term_2 = np.tile(term_2.reshape((1, n_loc, 3)), (n_time, 1, 1))
        oriented_term = term_1 * term_2

        return (front * (symmetric_term + oriented_term)).squeeze()

    def current_density(self, xyz):
        r"""Current density for the transient current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}`
        direction with dipole moment :math:`I ds`, this method computes
        the transient current density at the set of observation times for
        the gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988).
        For a transient electric current dipole oriented in the
        :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            {\bf e_e}(t) = \frac{Ids}{4\pi \sigma r^2} & \Bigg [ \Bigg ( \frac{x^2}{r^2}\mathbf{\hat x} + \frac{xy}{r^2}\mathbf{\hat y} + \frac{xz}{r^2}\mathbf{\hat z} \Bigg ) \Bigg ( 3 \, \textrm{erf}(\theta r) - \bigg ( \frac{4}{\sqrt{\pi}}\theta^3 r^3 \; ... \\
            & + \frac{6}{\sqrt{\pi}} \theta r \bigg ) e^{-\theta^2 r^2}  \Bigg ) - \Bigg ( \textrm{erf}(\theta r) - \bigg ( \frac{4}{\sqrt{\pi}} \theta^3 r^3 + \frac{2}{\sqrt{\pi}} \theta r \bigg ) e^{-\theta^2 r^2} \Bigg ) \mathbf{\hat x} \Bigg ]

        where

        .. math::
            \theta = \Bigg ( \frac{\mu\sigma}{4t} \Bigg )^{1/2}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_time, n_loc, 3) numpy.array of float
            Current density at all times for the gridded
            locations provided. Output array is squeezed when n_time and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an x-oriented electric dipole and plot the current density
        on the xz-plane that intercepts y=0.

        >>> from geoana.em.tdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> time = np.logspace(-6, -2, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[1., 0., 0.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     time, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the current density.

        >>> xyz = ndgrid(np.linspace(-10, 10, 20), np.array([0]), np.linspace(-10, 10, 20))
        >>> J = simulation.current_density(xyz)

        Finally, we plot the current density at the desired locations/times.

        >>> t_ind = 0
        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0::2], J[t_ind, :, 0::2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Current density at {} s'.format(time[t_ind]))

        """
        return self.sigma * self.electric_field(xyz)

    def magnetic_field(self, xyz):
        r"""Magnetic field for the transient current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}`
        direction with dipole moment :math:`I ds`, this method computes
        the transient magnetic field at the set of observation times for
        the gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988).
        For a transient electric current dipole oriented in the
        :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \mathbf{h}(t) = \frac{Ids}{4 \pi r^3} \bigg ( \textrm{erf}(\theta r) - \frac{2}{\sqrt{\pi}} \theta r \, e^{-\theta^2 r^2}  \bigg ) \big ( - z \, \mathbf{\hat y} + y \, \mathbf{\hat z}  \big )

        where

        .. math::
            \theta = \Bigg ( \frac{\mu\sigma}{4t} \Bigg )^{1/2}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_time, n_loc, 3) numpy.array of float
            Transient magnetic field at all times for the gridded
            locations provided. Output array is squeezed when n_time and/or
            n_loc = 1.

        Examples
        --------
        Here, we define a z-oriented electric dipole and plot the magnetic field
        on the xy-plane that intercepts z=0.

        >>> from geoana.em.tdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> time = np.logspace(-6, -2, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     time, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the magnetic field.

        >>> xyz = ndgrid(np.linspace(-10, 10, 20), np.linspace(-10, 10, 20), np.array([0]))
        >>> H = simulation.magnetic_field(xyz)

        Finally, we plot the magnetic field at the desired locations/times.

        >>> t_ind = 0
        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0:2], H[t_ind, :, 0:2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Y')
        >>> ax.set_title('Magnetic field at {} s'.format(time[t_ind]))

        """
        # dxyz = self.vector_distance(xyz)
        # r = self.distance(dxyz)
        # r = repeat_scalar(r)
        # thetar = self.theta * r

        # front = (
        #     self.current * self.length / (4 * np.pi * r**2) * (
        #         2 / np.sqrt(np.pi) * thetar * np.exp(-thetar**2) + erf(thetar)
        #     )
        # )

        # return - front * self.cross_orientation(xyz) / r

        dxyz = self.vector_distance(xyz)
        r = self.distance(dxyz)

        n_loc = len(r)
        n_time = len(self.time)

        theta_r = np.outer(self.theta, r)
        tile_r = np.outer(np.ones(n_time), r)

        term_1 = (
            self.current * self.length / (4 * np.pi * tile_r**2) * (
                2 / np.sqrt(np.pi) * theta_r * np.exp(-theta_r**2) + erf(theta_r)
            )
        ).reshape((n_time, n_loc, 1))

        r = repeat_scalar(r)
        term_2 = (self.cross_orientation(xyz) / r).reshape((1, n_loc, 3))

        return - (np.tile(term_1, (1, 1, 3)) * np.tile(term_2, (n_time, 1, 1))).squeeze()

    def magnetic_field_time_deriv(self, xyz):
        r"""Time-derivative of the magnetic field for the transient current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}`
        direction with dipole moment :math:`I ds`, this method computes
        the time-derivative of the transient magnetic field at the set of observation times for
        the gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988).
        For a transient electric current dipole oriented in the
        :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \frac{\partial \mathbf{h}}{\partial t} = - \frac{2 \, \theta^5 Ids}{\pi^{3/2} \mu \sigma} e^{-\theta^2 r^2} \big ( - z \, \mathbf{\hat y} + y \, \mathbf{\hat z}  \big )

        where

        .. math::
            \theta = \Bigg ( \frac{\mu\sigma}{4t} \Bigg )^{1/2}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_time, n_loc, 3) numpy.array of float
            Time-derivative of the transient magnetic field at all times for the gridded
            locations provided. Output array is squeezed when n_time and/or
            n_loc = 1.

        Examples
        --------
        Here, we define a z-oriented electric dipole and plot the time-derivative
        of the magnetic field on the xy-plane that intercepts z=0.

        >>> from geoana.em.tdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> time = np.logspace(-6, -2, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     time, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the dh/dt.

        >>> xyz = ndgrid(np.linspace(-10, 10, 20), np.linspace(-10, 10, 20), np.array([0]))
        >>> dHdt = simulation.magnetic_field_time_deriv(xyz)

        Finally, we plot dH/dt at the desired locations/times.

        >>> t_ind = 0
        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0:2], dHdt[t_ind, :, 0:2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Y')
        >>> ax.set_title('dH/dt at {} s'.format(time[t_ind]))

        """
        # dxyz = self.vector_distance(xyz)
        # r = self.distance(dxyz)
        # r = repeat_scalar(r)

        # front = (
        #     self.current * self.length * self.theta**3 * r /
        #     (2 * np.sqrt(np.pi)**3 * self.time)
        # )

        # return - front * self.cross_orientation(xyz) / r

        dxyz = self.vector_distance(xyz)
        r = self.distance(dxyz)

        n_loc = len(r)
        n_time = len(self.time)

        theta3_r = np.outer(self.theta**3, r)
        tile_t = np.outer(self.time, np.ones(n_loc))

        term_1 = (
            self.current * self.length * theta3_r /
            (2 * np.sqrt(np.pi)**3 * tile_t)
        ).reshape((n_time, n_loc, 1))

        r = repeat_scalar(r)
        term_2 = (self.cross_orientation(xyz) / r).reshape((1, n_loc, 3))

        return - (np.tile(term_1, (1, 1, 3)) * np.tile(term_2, (n_time, 1, 1))).squeeze()

    def magnetic_flux_density(self, xyz):
        r"""Magnetic flux density for the transient current dipole at a set of gridded locations.

        For an electric current dipole oriented in the :math:`\hat{u}`
        direction with dipole moment :math:`I ds`, this method computes
        the transient magnetic flux density at the set of observation times for
        the gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988).
        For a transient electric current dipole oriented in the
        :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \mathbf{b}(t) = \frac{\mu Ids}{4 \pi r^3} \bigg ( \textrm{erf}(\theta r) - \frac{2}{\sqrt{\pi}} \theta r \, e^{-\theta^2 r^2}  \bigg ) \big ( - z \, \mathbf{\hat y} + y \, \mathbf{\hat z}  \big )

        where

        .. math::
            \theta = \Bigg ( \frac{\mu\sigma}{4t} \Bigg )^{1/2}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_time, n_loc, 3) numpy.array of float
            Transient magnetic field at all times for the gridded
            locations provided. Output array is squeezed when n_time and/or
            n_loc = 1.

        Examples
        --------
        Here, we define a z-oriented electric dipole and plot the magnetic flux density
        on the xy-plane that intercepts z=0.

        >>> from geoana.em.tdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> time = np.logspace(-6, -2, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     time, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute the magnetic flux density.

        >>> xyz = ndgrid(np.linspace(-10, 10, 20), np.linspace(-10, 10, 20), np.array([0]))
        >>> B = simulation.magnetic_flux_density(xyz)

        Finally, we plot the magnetic flux density at the desired locations/times.

        >>> t_ind = 0
        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0:2], B[t_ind, :, 0:2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Y')
        >>> ax.set_title('Magnetic flux density at {} s'.format(time[t_ind]))

        """

        return self.mu * self.magnetic_field(xyz)

    def magnetic_flux_density_time_deriv(self, xyz):
        r"""Time-derivative of the magnetic flux density for the transient current dipole at a set of gridded locations.

        For a transient electric current dipole oriented in the :math:`\hat{u}`
        direction with dipole moment :math:`I ds`, this method computes
        the time-derivative of the time-derivative of the magnetic flux density at the set of observation times for
        the gridded xyz locations provided.

        The analytic solution is adapted from Ward and Hohmann (1988).
        For a transient electric current dipole oriented in the
        :math:`\hat{x}` direction, the solution at vector distance
        :math:`\mathbf{r}` from the current dipole is:

        .. math::
            \frac{\partial \mathbf{b}}{\partial t} = - \frac{2 \, \theta^5 Ids}{\pi^{3/2} \sigma} e^{-\theta^2 r^2} \big ( - z \, \mathbf{\hat y} + y \, \mathbf{\hat z}  \big )

        where

        .. math::
            \theta = \Bigg ( \frac{\mu\sigma}{4t} \Bigg )^{1/2}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n_time, n_loc, 3) numpy.array of float
            Time-derivative of the transient magnetic flux density at all times for the gridded
            locations provided. Output array is squeezed when n_time and/or
            n_loc = 1.

        Examples
        --------
        Here, we define a z-oriented electric dipole and plot the time-derivative of
        the magnetic flux density on the xy-plane that intercepts z=0.

        >>> from geoana.em.tdem import ElectricDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> time = np.logspace(-6, -2, 3)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> current = 1.
        >>> sigma = 1.0
        >>> simulation = ElectricDipoleWholeSpace(
        >>>     time, location=location, orientation=orientation,
        >>>     current=current, sigma=sigma
        >>> )

        Now we create a set of gridded locations and compute dB/dt.

        >>> xyz = ndgrid(np.linspace(-10, 10, 20), np.linspace(-10, 10, 20), np.array([0]))
        >>> dBdt = simulation.magnetic_flux_density_time_deriv(xyz)

        Finally, we plot the dB/dt at the desired locations/times.

        >>> t_ind = 0
        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0:2], dBdt[t_ind, :, 0:2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Y')
        >>> ax.set_title('dBdt at {} s'.format(time[t_ind]))

        """

        return self.mu * self.magnetic_field_time_deriv(xyz)



import numpy as np
from scipy.constants import mu_0, epsilon_0

###############################################################################
#                                                                             #
#                              Base Classes                                   #
#                                                                             #
###############################################################################


class BaseEM:
    """Base electromagnetics class.

    The base EM class stores the physical properties that are relevant to all problems
    which solve Maxwell's equations. The base EM class assumes that all physical properties
    (conductivity, magnetic permeability and dielectric permittivity) are homogeneously
    distributed within a wholespace and are non-dispersive. These properties are
    overwritten in child classes as necessary.

    Parameters
    ----------
    sigma : float, optional
        Electrical conductivity in S/m. Default is :math:`\\sigma` = 1 S/m
    mu : float, optional
        Magnetic permeability in H/m. Default is :math:`\\mu_0 = 4\\pi \\times 10^{-7}` H/m
    epsilon : float, optional
        Dielectric permittivity F/m. Default is :math:`\\epsilon_0 = 8.85 \\times 10^{-12}` F/m
    """

    def __init__(self, sigma=1.0, mu=mu_0, epsilon=epsilon_0, **kwargs):

        self.sigma = sigma
        self.mu = mu
        self.epsilon = epsilon
        super().__init__(**kwargs)


    @property
    def sigma(self):
        """Electrical conductivity in S/m

        Returns
        -------
        float
            Electrical conductivity in S/m
        """
        return self._sigma

    @sigma.setter
    def sigma(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"sigma must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("sigma must be greater than 0")

        self._sigma = value

    @property
    def mu(self):
        """Magnetic permeability in H/m

        Returns
        -------
        float
            Magnetic permeability in H/m
        """
        return self._mu

    @mu.setter
    def mu(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"mu must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("mu must be greater than 0")

        self._mu = value

    @property
    def epsilon(self):
        """Dielectric permittivity in F/m

        Returns
        -------
        float
            dielectric permittivity in F/m
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"epsilon must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("epsilon must be greater than 0")

        self._epsilon = value

class BaseDipole:
    """Base class for dipoles; namely the location and orientation.
    This class is inherited by both electric current and magnetic dipoles.

    Parameters
    ----------
    location: (3) array_like, optional
        Location of the dipole in 3D space. Default is (0,0,0)
    orientation : (3) array_like or {'X','Y','Z'}
        Orientation of the dipole. Can be defined using as an ``array_like`` of length 3,
        or by using one of {'X','Y','Z'} to define a unit dipole along the x, y or z direction.
        Default is 'X'.
    """

    def __init__(self, location=np.r_[0.,0.,0.], orientation='X', **kwargs):

        self.location = location
        self.orientation = orientation
        super().__init__(**kwargs)


    @property
    def location(self):
        """Location of the dipole

        Returns
        -------
        (3) numpy.ndarray of float
            xyz dipole location
        """
        return self._location

    @location.setter
    def location(self, vec):

        try:
            vec = np.asarray(vec, dtype=float)
        except:
            raise TypeError(f"location must be array_like of float, got {type(vec)}")
        vec = np.squeeze(vec)
        if vec.shape != (3, ):
            raise ValueError(
                f"location must be array_like with shape (3,), got {vec.shape}"
            )

        self._location = vec

    @property
    def orientation(self):
        """Orientation of the dipole as a normalized vector

        Returns
        -------
        (3) numpy.ndarray of float or str in {'X','Y','Z'}
            dipole orientation, normalized to unit magnitude
        """
        return self._orientation

    @orientation.setter
    def orientation(self, var):

        if isinstance(var, str):
            if var.upper() == 'X':
                var = np.r_[1., 0., 0.]
            elif var.upper() == 'Y':
                var = np.r_[0., 1., 0.]
            elif var.upper() == 'Z':
                var = np.r_[0., 0., 1.]
        else:
            try:
                var = np.asarray(var, dtype=float)
            except:
                raise TypeError(
                    f"orientation must be str or array_like, got {type(var)}"
                )
            var = np.squeeze(var)
            if var.shape != (3, ):
                raise ValueError(
                    f"orientation must be array_like with shape (3,), got {len(var)}"
                )

            # Normalize the orientation
            var /= np.linalg.norm(var)

        self._orientation = var

    def vector_distance(self, xyz):
        r"""Vector distance to a set of gridded xyz locations.

        Where :math:`\mathbf{p}` is the location of the source and :math:`\mathbf{q}`
        is a point in 3D space, this method returns the vector distance:

        .. math::
            \mathbf{v} = \mathbf{q} - \mathbf{p}

        for all locations :math:`\mathbf{q}` supplied in the inputed argument `xyz`.
        For dipoles, :math:`\mathbf{p}` is the dipole location. For circular loops,
        :math:`\mathbf{p}` defines the center location for the loop.

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n, 3) numpy.ndarray
            Vector distances in the x, y and z directions
        """
        return xyz - self.location

    def distance(self, xyz):
        r"""Scalar distance from dipole to a set of gridded xyz locations

        Where :math:`\mathbf{p}` is the location of the source and :math:`\mathbf{q}`
        is a point in 3D space, this method returns the scalar distance:

        .. math::
            d = \sqrt{(q_x - p_x)^2 + (q_y - p_y)^2 + (q_z - p_z)^2}

        for all locations :math:`\mathbf{q}` supplied in the input argument `xyz`.
        For dipoles, :math:`\mathbf{p}` is the dipole location. For circular loops,
        :math:`\mathbf{p}` defines the center location for the loop.

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            Gridded xyz locations

        Returns
        -------
        (n) numpy.ndarray
            Scalar distances from dipole to xyz locations
        """
        return np.linalg.norm(xyz - self.location, axis=-1)

    def dot_orientation(self, vecs):
        r"""Dot product between the orientation of the source and a gridded set of vectors.

        Where :math:`\mathbf{p}` is the vector defining the dipole moment direction and
        :math:`\mathbf{v}` is a 3D vector, this method returns the dot product:

        .. math::
            \mathbf{p} \cdot \mathbf{v}

        for all vectors :math:`\mathbf{v}` supplied in the input argument ``vecs``.

        Parameters
        ----------
        vecs : (n, 3) numpy.ndarray
            A set of 3D vectors

        Returns
        -------
        (n) numpy.ndarray
            Dot product between the dipole moment direction (orientation) and each vector supplied.
        """
        return np.dot(vecs, self.orientation)

    def cross_orientation(self, xyz):
        r"""Cross products between a gridded set of vectors and the orientation of the source.

        Where :math:`\mathbf{p}` is the vector defining the orientation of the source and
        :math:`\mathbf{v}` is a 3D vector, this method returns the cross product:

        .. math::
            \mathbf{v} \times \mathbf{p}


        for all vectors :math:`\mathbf{v}` supplied in the input argument ``vecs``.

        Parameters
        ----------
        vecs : (n, 3) numpy.ndarray
            A set of 3D vectors

        Returns
        -------
        (n) numpy.ndarray
            Cross product between each vector supplied and the source orientation.

        """
        return np.cross(xyz, self.orientation)


class BaseElectricDipole(BaseDipole):
    r"""Base class for electric current dipoles.

    The ``BaseElectricDipole`` class defines the basic properties for electric
    current dipoles. Where :math:`I` is the current and :math:`d\mathbf{s}` defines
    the direction and length of the dipole, the current dipole is defined as:

    .. math::
        \mathbf{j_s}(\mathbf{r}) = Id \mathbf{s} \, \delta(|\mathbf{r} - \mathbf{r_s}|)

    where :math:`\mathbf{r_s}` is the location of the source.
    For more, visit `EM GeoSci <https://em.geosci.xyz/content/maxwell1_fundamentals/dipole_sources_in_homogeneous_media/electric_dipole_definition/index.html>`__ .

    Parameters
    ----------
    length : float, int
        Length of the electric current dipole (m). Default is 1 m.
    current : float, int
        Current of the electric current dipole (A). Default is 1 A.
    """

    def __init__(self, length=1.0, current=1.0, **kwargs):

        self.length = length
        self.current = current

        super().__init__(**kwargs)

    @property
    def length(self):
        """Length of the electric current dipole (m)

        Returns
        -------
        float
            Length of the electric current dipole (m)
        """
        return self._length

    @length.setter
    def length(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"length must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("length must be greater than 0")

        self._length = value

    @property
    def current(self):
        """Current in the electric current dipole (A)

        Returns
        -------
        float
            Current in the electric current dipole (A)
        """
        return self._current

    @current.setter
    def current(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"current must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("current must be greater than 0")

        self._current = value


class BaseMagneticDipole(BaseDipole):
    r"""Base class for magnetic dipoles.

    The ``BaseMagneticDipole`` class defines the basic properties for magnetic dipoles.
    Where :math:`d\mathbf{m}` defines the dipole moment, the magnetic dipole is defined as:

    .. math::
        \mathbf{s_m}(\mathbf{r}) = \mathbf{m} \, \delta(|\mathbf{r} - \mathbf{r_s}|)

    where :math:`\mathbf{r_s}` is the location of the source.
    For more, visit `EM GeoSci <https://em.geosci.xyz/content/maxwell1_fundamentals/dipole_sources_in_homogeneous_media/index.html>`__ .


    Parameters
    ----------
    moment : float, optional
        Amplitude of the dipole moment for the magnetic dipole (:math:`A/m^2`).
        Default is 1 :math:`A/m^2`.
    """

    def __init__(self, moment=1.0, **kwargs):

        self.moment = moment
        super().__init__(**kwargs)

    @property
    def moment(self):
        """Amplitude of the dipole moment of the magnetic dipole (:math:`A/m^2`)

        Returns
        -------
        float
            Amplitude of the dipole moment of the magnetic dipole (:math:`A/m^2`)
        """
        return self._moment

    @moment.setter
    def moment(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"moment must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("moment must be greater than 0")

        self._moment = value


class BaseLineCurrent:
    """Base class for connected segments of current-carrying wire.

    Parameters
    ----------
    nodes : (n, 3) array_like
        Nodes defining the segments of current-carrying wire segments. For an
        inductive source, you must close the loop.
    """

    def __init__(self, nodes, current=1.0, **kwargs):

        self.nodes = nodes
        self.current = current
        super().__init__(**kwargs)

    @property
    def nodes(self):
        """Nodes defining the segments of current-carrying wire segments;
        close the loop if inductive.

        Returns
        -------
        (n_nodes, 3) numpy.ndarray of float
            node locations
        """
        return self._nodes

    @nodes.setter
    def nodes(self, xyz):

        try:
            xyz = np.asarray(xyz, dtype=np.float64)
        except:
            raise TypeError(f"nodes must be array_like, got {type(xyz)}")
        xyz = np.atleast_2d(xyz)

        if (xyz.shape[1] != 3) | (xyz.ndim != 2):
            raise ValueError(
                f"nodes must be array_like with shape (n, 3), got {xyz.shape}"
            )

        self._nodes = xyz

    @property
    def n_segments(self):
        """Number of wire segments

        Returns
        -------
        int
            number of wire segments

        """
        return self.nodes.shape[0] - 1


    @property
    def current(self):
        """Current in wire (A)

        Returns
        -------
        float
            Current in the wire (A)
        """
        return self._current

    @current.setter
    def current(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"current must be a number, got {type(value)}")

        self._current = value



"""
=================================================
Plotting Utilities (:mod:`geoana.plotting_utils`)
=================================================
.. currentmodule:: geoana.plotting_utils

The ``geoana.plotting_utils`` module provides some functions for plotting
data computed with geoana.

.. autosummary::
  :toctree: generated/

  plot2Ddata

"""

import numpy as np
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator
import matplotlib.pyplot as plt
from matplotlib import colors
import warnings


def plot2Ddata(
    xyz,
    data,
    vec=False,
    nx=100,
    ny=100,
    ax=None,
    mask=None,
    level=False,
    figname=None,
    ncontour=10,
    dataloc=False,
    contourOpts={},
    levelOpts={},
    streamplotOpts={},
    scale="linear",
    clim=None,
    method="linear",
    shade=False,
    shade_ncontour=100,
    shade_azimuth=-45.0,
    shade_angle_altitude=45.0,
    shadeOpts={},
):
    """
    Take unstructured xy points, interpolate, then plot in 2D

    Parameters
    ----------
    xyz : (n, dim) numpy.ndarray
        Data locations. Only the first two columns are used if an (n, 3)
        array is entered.
    data : (n) or (n, 2) np.ndarray
        Data values. Either scalar or 2D vector.
    vec : bool
        Default = ``False``. Plot vector data. If ``False``, the ``data`` input argument is 
        scalar-valued. If ``True``, the ``data`` input arument is vector.
    nx : int
        number of x grid locations
    ny : int
        number of y grid locations
    ax : matplotlib.axes
        An axes object
    mask : np.ndarray of bool
        Masking array
    level: bool
        To plot or not
    figname : string
        figure name
    ncontour : int
        number of :meth:`matplotlib.pyplot.contourf` contours
    dataloc : bool
        If ``True``, plot the data locations. If ``False``, do not. Default = ``False``
    contourOpts : dict
        Contour plot options. Dictionary of :meth:`matplotlib.pyplot.contourf` options
    levelOpts : dict
        Level options. Dictionary of :meth:`matplotlib.pyplot.contour` options
    clim : (2) np.ndarray
        Colorbar limits
    method : str {'linear', 'nearest'}
        interpolation method from ``xyz`` locations and the gridded locations for
        the contour plot
    shade : bool
        Add shading to the plot. Default = ``False``
    shade_ncontour : int
        number of :meth:`matplotlib.pyplot.contourf` contours for the shading
    shade_azimuth : float
        azimuth for the light source in shading
    shade_angle_altitude : float
        angle altitude for the light source in shading
    shapeOpts : dict
        Dictionary of :meth:`matplotlib.pyplot.contourf` options

    Returns
    -------
    :meth:`matplotlib.pyplot.contourf`
        Contour object
    matplotlib.axes
        Axes

    """

    # Error checking and set vmin, vmax
    vlimits = [None, None]

    if clim is not None:
        vlimits = [np.min(clim), np.max(clim)]

    for i, key in enumerate(["vmin", "vmax"]):
        if key in contourOpts.keys():
            if vlimits[i] is None:
                vlimits[i] = contourOpts.pop(key)
            else:
                if not np.isclose(contourOpts[key], vlimits[i]):
                    raise Exception(
                        "The values provided in the colorbar limit, clim {} "
                        "does not match the value of {} provided in the "
                        "contourOpts: {}. Only one value should be provided or "
                        "the two values must be equal.".format(
                            vlimits[i], key, contourOpts[key]
                        )
                    )
                contourOpts.pop(key)
    vmin, vmax = vlimits[0], vlimits[1]

    # create a figure if it doesn't exist
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    # interpolate data to grid locations
    xmin, xmax = xyz[:, 0].min(), xyz[:, 0].max()
    ymin, ymax = xyz[:, 1].min(), xyz[:, 1].max()
    x = np.linspace(xmin, xmax, nx)
    y = np.linspace(ymin, ymax, ny)
    X, Y = np.meshgrid(x, y)
    xy = np.c_[X.flatten(), Y.flatten()]

    if vec is False:
        if method == "nearest":
            F = NearestNDInterpolator(xyz[:, :2], data)
        else:
            F = LinearNDInterpolator(xyz[:, :2], data)
        DATA = F(xy)
        DATA = DATA.reshape(X.shape)

        # Levels definitions
        dataselection = np.logical_and(~np.isnan(DATA), np.abs(DATA) != np.inf)
        if scale == "log":
            DATA = np.abs(DATA)

        # set vmin, vmax if they are not already set
        vmin = DATA[dataselection].min() if vmin is None else vmin
        vmax = DATA[dataselection].max() if vmax is None else vmax

        if scale == "log":
            levels = np.logspace(np.log10(vmin), np.log10(vmax), ncontour + 1)
            norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        else:
            levels = np.linspace(vmin, vmax, ncontour + 1)
            norm = colors.Normalize(vmin=vmin, vmax=vmax)

        if mask is not None:
            Fmask = NearestNDInterpolator(xyz[:, :2], mask)
            MASK = Fmask(xy)
            MASK = MASK.reshape(X.shape)
            DATA = np.ma.masked_array(DATA, mask=MASK)

        contourOpts = {"levels": levels, "norm": norm, "zorder": 1, **contourOpts}
        cont = ax.contourf(X, Y, DATA, **contourOpts)

        if level:
            levelOpts = {"levels": levels, "zorder": 3, **levelOpts}
            CS = ax.contour(X, Y, DATA, **levelOpts)

    else:
        # Assume size of data is (N,2)
        datax = data[:, 0]
        datay = data[:, 1]
        if method == "nearest":
            Fx = NearestNDInterpolator(xyz[:, :2], datax)
            Fy = NearestNDInterpolator(xyz[:, :2], datay)
        else:
            Fx = LinearNDInterpolator(xyz[:, :2], datax)
            Fy = LinearNDInterpolator(xyz[:, :2], datay)
        DATAx = Fx(xy)
        DATAy = Fy(xy)
        DATA = np.sqrt(DATAx ** 2 + DATAy ** 2).reshape(X.shape)
        DATAx = DATAx.reshape(X.shape)
        DATAy = DATAy.reshape(X.shape)
        if scale == "log":
            DATA = np.abs(DATA)

        # Levels definitions
        dataselection = np.logical_and(~np.isnan(DATA), np.abs(DATA) != np.inf)

        # set vmin, vmax
        vmin = DATA[dataselection].min() if vmin is None else vmin
        vmax = DATA[dataselection].max() if vmax is None else vmax

        if scale == "log":
            levels = np.logspace(np.log10(vmin), np.log10(vmax), ncontour + 1)
            norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        else:
            levels = np.linspace(vmin, vmax, ncontour + 1)
            norm = colors.Normalize(vmin=vmin, vmax=vmax)

        if mask is not None:
            Fmask = NearestNDInterpolator(xyz[:, :2], mask)
            MASK = Fmask(xy)
            MASK = MASK.reshape(X.shape)
            DATA = np.ma.masked_array(DATA, mask=MASK)

        contourOpts = {"levels": levels, "norm": norm, "zorder": 1, **contourOpts}
        cont = ax.contourf(X, Y, DATA, **contourOpts)

        streamplotOpts = {"zorder": 4, "color": "w", **streamplotOpts}
        ax.streamplot(X, Y, DATAx, DATAy, **streamplotOpts)

        if level:
            levelOpts = {"levels": levels, "zorder": 3, **levelOpts}
            CS = ax.contour(X, Y, DATA, levels=levels, zorder=3, **levelOpts)

    if shade:

        def hillshade(array, azimuth, angle_altitude):
            """
            coded copied from https://www.neonscience.org/create-hillshade-py
            """
            azimuth = 360.0 - azimuth
            x, y = np.gradient(array)
            slope = np.pi / 2.0 - np.arctan(np.sqrt(x * x + y * y))
            aspect = np.arctan2(-x, y)
            azimuthrad = azimuth * np.pi / 180.0
            altituderad = angle_altitude * np.pi / 180.0
            shaded = np.sin(altituderad) * np.sin(slope) + np.cos(altituderad) * np.cos(
                slope
            ) * np.cos((azimuthrad - np.pi / 2.0) - aspect)
            return 255 * (shaded + 1) / 2

        shadeOpts = {
            "cmap": "Greys",
            "alpha": 0.35,
            "antialiased": True,
            "zorder": 2,
            **shadeOpts,
        }

        ax.contourf(
            X,
            Y,
            hillshade(DATA, shade_azimuth, shade_angle_altitude),
            shade_ncontour,
            **shadeOpts
        )

    if dataloc:
        ax.plot(xyz[:, 0], xyz[:, 1], "k.", ms=2)
    ax.set_aspect("equal", adjustable="box")
    if figname:
        plt.axis("off")
        fig.savefig(figname, dpi=200)
    if level:
        return cont, ax, CS
    else:
        return cont, ax



import numpy as np
from scipy.constants import mu_0
from scipy.special import iv, kv
from geoana.em.base import BaseMagneticDipole
from geoana.em.fdem.base import BaseFDEM


class MagneticDipoleHalfSpace(BaseFDEM, BaseMagneticDipole):
    r"""Class for a harmonic magnetic dipole in a wholespace.
    """

    def __init__(self, frequency, **kwargs):
        super().__init__(frequency=frequency, **kwargs)
        self._check_is_valid_location()

    def _check_is_valid_location(self):
        if self.location[2] != 0.0:
            raise ValueError("Source must be at the surface of the earth (i.e. z=0.0)")


    def magnetic_field(self, xy, field="secondary"):
        """Magnetic field due to a magnetic dipole over a half space

        The analytic expression is only valid for a source and receiver at the
        surface of the earth. For arbitrary source and receiver locations above
        the earth, use the layered solution.

        Parameters
        ----------
        xy : (..., 2) numpy.ndarray
            receiver locations of shape
        field : ("secondary", "total")
            Flag for the type of field to return.

        Returns
        -------
        (n_freq, ..., 3) numpy.array of complex
            Magnetic field at all frequencies for the gridded
            locations provided. Output array is squeezed when n_freq and/or
            n_loc = 1.

        Examples
        --------
        Here, we define an z-oriented magnetic dipole at (0, 0, 0) and plot
        the secondary magnetic field at multiple frequencies at (5, 0, 0).

        >>> from geoana.em.fdem import MagneticDipoleHalfSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the electric current dipole.

        >>> frequency = np.logspace(2, 6, 41)
        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> sigma = 1.0
        >>> simulation = MagneticDipoleHalfSpace(
        >>>     frequency, location=location, orientation=orientation,
        >>>     moment=moment, sigma=sigma
        >>> )

        Now we define the receiver location and plot the secondary field.

        >>> xyz = np.c_[5, 0, 0]
        >>> H = simulation.magnetic_field(xyz, field='secondary')

        Finally, we plot the real and imaginary components of the magnetic field.

        >>> fig = plt.figure(figsize=(6, 4))
        >>> ax1 = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> ax1.semilogx(frequency, np.real(H[:, 2]), 'r', lw=2)
        >>> ax1.semilogx(frequency, np.imag(H[:, 2]), 'r--', lw=2)
        >>> ax1.set_xlabel('Frequency (Hz)')
        >>> ax1.set_ylabel('Secondary field (H/m)')
        >>> ax1.grid()
        >>> ax1.autoscale(tight=True)
        >>> ax1.legend(['real', 'imaginary'])


        """
        f = self.frequency
        sig = self.sigma_hat
        w = 2*np.pi*f
        k = np.sqrt(-1j*w*mu_0*sig)  # K shape is (n_freq, )

        dxy = xy[..., :2] - self.location[:2]
        r = np.linalg.norm(dxy, axis=-1)
        x = dxy[..., 0]
        y = dxy[..., 1]

        for dim in range(r.ndim):
            k = k[:, None]

        em_x = em_y = em_z = 0
        src_x, src_y, src_z = self.orientation

        alpha = 1j*k*r/2.  # (n_freq, ...) * (...)
        IK1 = iv(1, alpha)*kv(1, alpha)
        IK2 = iv(2, alpha)*kv(2, alpha)

        if src_z != 0.0:
            # Z component of source
            em_z += src_z*2.0/(k**2*r**5)*(9-(9+9*1j*k*r-4*k**2*r**2-1j*k**3*r**3)*np.exp(-1j*k*r))
            Hr = (k**2/r)*(IK1 - IK2)
            angle = np.arctan2(y, x)
            em_x += src_z*np.cos(angle)*Hr
            em_y += src_z*np.sin(angle)*Hr

        if src_x != 0.0 or src_y != 0.0:
            # X component of source
            phi = 2/(k**2*r**4)*(3 + k**2*r**2 - (3 + 3j*k*r - k**2*r**2)*np.exp(-1j*k*r))
            dphi_dr = 2/(k**2*r**5)*(-2*k**2*r**2 - 12 + (-1j*k**3*r**3 - 5*k**2*r**2 + 12j*k*r + 12)*np.exp(-1j*k*r))
            if src_x != 0.0:
                em_x += src_x*(-1.0/r**3)*(y**2*phi + x**2*r*dphi_dr)
                em_y += src_x*(1.0/r**3)*x*y*(phi - r*dphi_dr)
                em_z -= src_x*(k**2*x/r**2)*(IK1 - IK2)

            # Y component of source
            if src_y != 0.0:
                em_x += src_y*(1.0/r**3)*x*y*(phi - r*dphi_dr)
                em_y += src_y*(-1.0/r**3)*(x**2*phi + y**2*r*dphi_dr)
                em_z -= src_y*(k**2*y/r**2)*(IK1 - IK2)

        if field == "secondary":
            # subtract out primary field from above
            mdotr = src_x*x + src_y*y # + m[2]*(z=0)

            em_x -= 3*x*mdotr/r**5 - src_x/r**3
            em_y -= 3*y*mdotr/r**5 - src_y/r**3
            em_z -= -src_z/r**3 # + 3*(z=0)*mdotr/r**5

        return self.moment/(4*np.pi)*np.stack((em_x, em_y, em_z), axis=-1).squeeze()



from geoana.em.base import BaseEM
import numpy as np
from scipy.constants import mu_0, epsilon_0

###############################################################################
#                                                                             #
#                           Utility Functions                                 #
#                                                                             #
###############################################################################


def omega(frequency):
    """Compute angular frequencies.

    For input frequencies :math:`f`, this function returns the
    corresponding angular frequencies:

    .. math::
        \\omega = 2 \\pi f

    Parameters
    ----------
    frequency : float, numpy.ndarray
        frequency or frequencies (Hz)

    Returns
    -------
    float, numpy.ndarray
        Angular frequency or frequencies in rad/s

    """
    return 2*np.pi*frequency


def wavenumber(
    frequency, sigma, mu=mu_0, epsilon=epsilon_0, quasistatic=False
):
    """Compute wavenumber for an electromagnetic wave in a homogeneous isotropic medium.

    Where :math:`f` is the frequency of the EM wave in Hertz, :math:`\\sigma` is the
    electrical conductivity in S/m, :math:`\\mu` is the magnetic permeability in
    H/m and :math:`\\varepsilon` is the dielectric permittivity in F/m, the
    wavenumber is given by:

    .. math::
        k = \\sqrt{\\omega^2 \\mu \\varepsilon - i \\omega \\mu \\sigma}

    where

    .. math::
        \\omega = 2 \\pi f

    Parameters
    ----------
    frequency : float, numpy.ndarray
        frequency or frequencies at which you are computing the wavenumber (Hz)
    sigma : float
        electrical conductivity (S/m).
    mu : float (optional)
        magnetic permeability (H/m). Default: :math:`\\mu_0 = 4\\pi \\times 10^{-7}` H/m.
    epsilon : float (optional)
        dielectric permittivity (F/m). Default: :math:`\\epsilon_0 = 8.85 \\times 10^{-12}` F/m.
    quasistatic : bool (optional)
        use the quasi-static assumption; i.e. ignore the dielectric term. Default: False

    Returns
    -------
    complex, (n_freq) numpy.ndarray of complex
        Wavenumber for all frequencies provided

    """
    w = omega(frequency)
    if quasistatic is True:
        return np.sqrt(-1j * w * mu * sigma)
    return np.sqrt(w**2 * mu * epsilon - 1j * w * mu * sigma)


def skin_depth(frequency, sigma, mu=mu_0, epsilon=epsilon_0, quasistatic=True):
    r"""Compute skin depth for an electromagnetic wave in a homogeneous isotropic medium.

    The skin depth is the propagation distance at which an EM planewave has decayed by a factor
    of :math:`1/e`. For a homogeneous medium with non-dispersive electrical conductivity
    :math:`\sigma`, magnetic permeability :math:`\mu` and dielectric permittivity
    :math:`\varepsilon`, the skin depth for a wave at frequency :math:`f` is given by:

    .. math::
        \delta = \frac{1}{\omega} \Bigg (\frac{\mu \varepsilon}{2} \bigg [ \bigg (
        1 + \frac{\sigma^2}{\omega^2 \varepsilon^2} \bigg )^{1/2} - 1 \bigg ] \Bigg )^{1/2}

    where :math:`\omega` is the angular frequency:

    .. math::
        \omega = 2 \pi f

    For the quasistatic approximation, dielectric permittivity is ignore and the
    skin depth simplifies to:

    .. math::
        \delta = \sqrt{\frac{2}{\omega \sigma \mu}}

    Parameters
    ----------
    frequency : float, numpy.ndarray
        frequency or frequencies (Hz)
    sigma : float
        electrical conductivity (S/m)
    mu : float (optional)
        magnetic permeability (H/m). Default: :math:`\mu_0 = 4\pi \times 10^{-7}` H/m
    epsilon : float (optional)
        dielectric permittivity (F/m). Default: :math:`\epsilon_0 = 8.85 \times 10^{-12}` F/m.
    quasistatic : bool (optional)
        If ``True``, the quasistatic approximation for the skin depth is computed.

    Returns
    -------
    float, (n_frequencies) numpy.ndarray
        Skin depth for all frequencies provided

    """
    w = omega(frequency)
    if quasistatic:
        return np.sqrt(2./(w*sigma*mu))
    return np.sqrt((mu*epsilon/2) * (np.sqrt(1 + sigma**2/(w*epsilon)**2) - 1)) / w


def sigma_hat(frequency, sigma, epsilon=epsilon_0, quasistatic=False):
    """Compute the electrical conductivity including electric displacement.

    Where :math:`\\sigma` is the electrical conductivity, :math:`\\varepsilon` is the
    dielectric permittivity and :math:`\\omega` is the angular frequency, this function
    returns:

    .. math::

        \\hat{\\sigma} = \\sigma + i \\omega \\varepsilon

    Parameters
    ----------
    frequency : float, numpy.ndarray
        frequency or frequencies (Hz)
    sigma : float
        electrical conductivity (S/m)
    epsilon : float (optional)
        dielectric permittivity (F/m). Default: :math:`\\epsilon_0 = 8.85 \\times 10^{-12}` F/m
    quasistatic : bool (optional)
        use the quasi-static assumption; i.e. ignore the dielectric term. Default: ``False``

    Returns
    -------
    complex, (n_frequencies) numpy.ndarray
        Conductivity with electric displacement for all frequencies provided

    """
    if quasistatic is True:
        return sigma
    return sigma + 1j*omega(frequency)*epsilon


###############################################################################
#                                                                             #
#                                  Classes                                    #
#                                                                             #
###############################################################################

class BaseFDEM(BaseEM):
    r"""
    Base frequency domain electromagnetic class

    The base FDEM class is contructed to define rudimentary properties and methods
    for frequency-domain electromagnetics.

    Parameters
    ----------
    frequency : int, float or (n_freq) numpy.ndarray
        Frequency or frequencies used for all computations in Hz.
    quasistatic : bool
        If ``True``, we assume the quasistatic approximation and dielectric permittivity
        is neglected in all computations. Default is ``False``.
    """

    def __init__(self, frequency, quasistatic=False, **kwargs):

        self.frequency = frequency
        self.quasistatic = quasistatic
        super().__init__(**kwargs)

    @property
    def frequency(self):
        """Frequency (Hz) used for all computations

        Returns
        -------
        numpy.ndarray
            Frequency (or frequencies) in Hz used for all computations
        """
        return self._frequency

    @frequency.setter
    def frequency(self, value):

        # Ensure float or numpy array of float
        try:
            value = np.asarray(value, dtype=float)
        except:
            raise TypeError(f"frequencies are not a valid type")
        value = np.atleast_1d(value)

        # Enforce positivity and dimensions
        if (value < 0.).any():
            raise ValueError("All frequencies must be greater than 0")
        if value.ndim > 1:
            raise TypeError(f"frequencies must be ('*') array")

        self._frequency = value

    @property
    def omega(self):
        """Angular frequency

        Where :math:`f` is the frequency assigned when instantiating the base EM class,
        this method returns the corresponding angular frequency:

        .. math::
            \\omega = 2\\pi f

        Returns
        -------
        float
            Angular frequency (rad/s)
        """
        return omega(self.frequency)

    @property
    def sigma_hat(self):
        """Conductivity including electric displacement term.

        Where :math:`\\sigma` is the electrical conductivity, :math:`\\varepsilon` is
        the dielectric permittivity and :math:`\\omega = 2\\pi f` is the angular
        frequency, this property returns:

        .. math::
            \\hat{\\sigma} = \\sigma + i \\omega \\varepsilon

        Returns
        -------
        complex
            Electrical conductivity including electric displacement. Returns the electrical
            conductivity :math:`\\sigma` if the property `quasistatic` is ``True``.

        """
        sigma = sigma_hat(
            self.frequency, self.sigma, epsilon=self.epsilon,
            quasistatic=self.quasistatic
        )
        if (np.imag(sigma) == 0).all():
            sigma = np.real(sigma)
        return sigma

    @property
    def wavenumber(self):
        r"""Wavenumber for an electromagnetic planewave in a homogenous isotropic medium.

        Where :math:`\sigma` is the electrical conductivity, :math:`\mu` is the magnetic
        permeability, :math:`\varepsilon` is the dielectric permittivity and
        :math:`\omega = 2\pi f` is the angular frequency, the wavenumber is:

        .. math::
            k = \sqrt{\omega^2 \mu \varepsilon - i \omega \mu \sigma}

        In the quasistatic regime, this expression simplifies to:

        .. math::
            k = \sqrt{-i \omega \mu \sigma}

        Returns
        -------
        complex
            Wavenumber for an electromagnetic planewave in a homogeneous isotropic medium.
            Returns the quasistatic approximation if the property `quasistatic` of the
            class instance is ``True``.

        """
        return wavenumber(
            self.frequency, self.sigma, mu=self.mu, epsilon=self.epsilon, quasistatic=self.quasistatic
        )

    @property
    def skin_depth(self):
        r"""Returns the skin depth for an electromagnetic wave in a homogeneous isotropic medium.

        The skin depth is the propagation distance at which an EM planewave has decayed by a factor
        of :math:`1/e`. For a homogeneous medium with non-dispersive electrical conductivity
        :math:`\sigma`, magnetic permeability :math:`\mu` and dielectric permittivity
        :math:`\varepsilon`, the skin depth for a wave at frequency :math:`f` is given by:

        .. math::
            \delta = \frac{1}{\omega} \Bigg (\frac{\mu \varepsilon}{2} \bigg [ \bigg (
            1 + \frac{\sigma^2}{\omega^2 \varepsilon^2} \bigg )^{1/2} - 1 \bigg ] \Bigg )^{1/2}

        where :math:`\omega` is the angular frequency:

        .. math::
            \omega = 2 \pi f

        For the quasistatic approximation, dielectric permittivity is ignore and the
        skin depth simplifies to:

        .. math::

            \delta = \sqrt{\frac{2}{\omega \sigma \mu}}

        Returns
        -------
        float
            Skin depth for the EM planewave. Returns the quasistatic approximation
            if the property `quasistatic` of the class instance is ``True``.

        """
        return skin_depth(
            self.frequency, self.sigma, mu=self.mu, epsilon=self.epsilon, quasistatic=self.quasistatic
        )



import numpy as np
from scipy.constants import epsilon_0


class ElectrostaticSphere:
    """Class for electrostatic solutions for a sphere in a wholespace.

    The ``ElectrostaticSphere`` class is used to analytically compute the electric
    potentials, fields, currents and charge densities for a sphere in a wholespace.
    For this class, we assume a homogeneous primary electric field along the
    :math:`\\hat{x}` direction.

    Parameters
    ----------
    radius : float
        radius of sphere (m).
    sigma_sphere : float
        conductivity of target sphere (S/m)
    sigma_background : float
        background conductivity (S/m)
    amplitude : float, optional
        amplitude of primary electric field along the :math:`\\hat{x}` direction (V/m).
        Default is 1.
    location : (3) array_like, optional
        Center of the sphere. Defaults is (0, 0, 0).
    """

    def __init__(
        self, radius, sigma_sphere, sigma_background, amplitude=1.0, location=np.r_[0.,0.,0.]
    ):

        self.radius = radius
        self.sigma_sphere = sigma_sphere
        self.sigma_background = sigma_background
        self.amplitude = amplitude
        self.location = location

    @property
    def sigma_sphere(self):
        """Electrical conductivity of the sphere in S/m

        Returns
        -------
        float
            Electrical conductivity of the sphere in S/m
        """
        return self._sigma_sphere

    @sigma_sphere.setter
    def sigma_sphere(self, item):
        item = float(item)
        if item <= 0.0:
            raise ValueError('Conductiviy must be positive')
        self._sigma_sphere = item

    @property
    def sigma_background(self):
        """Electrical conductivity of the background in S/m

        Returns
        -------
        float
            Electrical conductivity of the background in S/m
        """
        return self._sigma_background

    @sigma_background.setter
    def sigma_background(self, item):
        item = float(item)
        if item <= 0.0:
            raise ValueError('Conductiviy must be positive')
        self._sigma_background = item

    @property
    def radius(self):
        """Radius of the sphere in meters

        Returns
        -------
        float
            Radius of the sphere in meters
        """
        return self._radius

    @radius.setter
    def radius(self, item):
        item = float(item)
        if item < 0.0:
            raise ValueError('radius must be non-negative')
        self._radius = item

    @property
    def amplitude(self):
        """Amplitude of the primary current density along the x-direction.

        Returns
        -------
        float
            Amplitude of the primary current density along the x-direction in :math:`A/m^2`.
        """
        return self._amplitude

    @amplitude.setter
    def amplitude(self, item):
        self._amplitude = float(item)

    @property
    def location(self):
        """Center of the sphere

        Returns
        -------
        (3) numpy.ndarray of float
            Center of the sphere. Default = np.r_[0,0,0]
        """
        return self._location

    @location.setter
    def location(self, vec):

        try:
            vec = np.atleast_1d(vec).astype(float)
        except:
            raise TypeError(f"location must be array_like, got {type(vec)}")

        if len(vec) != 3:
            raise ValueError(
                f"location must be array_like with shape (3,), got {len(vec)}"
            )

        self._location = vec

    def _check_XYZ(self, XYZ):
        if len(XYZ) == 3:
            x, y, z = XYZ
            x = np.asarray(x, dtype=float)
            y = np.asarray(y, dtype=float)
            z = np.asarray(z, dtype=float)
        elif isinstance(XYZ, np.ndarray) and XYZ.shape[-1] == 3:
            x, y, z = XYZ[..., 0], XYZ[..., 1], XYZ[..., 2]
        else:
            raise TypeError(
                "XYZ must be either a length three tuple of each dimension, "
                "or a numpy.ndarray of shape (..., 3)."
                )
        if not (x.shape == y.shape and x.shape == z.shape):
            raise ValueError(
                "x, y, z must all have the same shape"
            )
        return x, y, z

    def potential(self, XYZ, field='all'):
        """Compute the electric potential.

        Parameters
        ----------
        XYZ : (3, ) tuple of np.ndarray or (..., 3) np.ndarray
            locations to evaluate at. If a tuple, all
            the numpy arrays must be the same shape.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Vt, Vp, Vs : (..., ) np.ndarray
            If field == "all"
        V : (..., ) np.ndarray
            If only requesting a single field.
        """
        sig0 = self.sigma_background
        sig1 = self.sigma_sphere
        E0 = self.amplitude
        sig_cur = (sig1 - sig0) / (sig1 + 2 * sig0)
        x0, y0, z0 = self.location
        x, y, z = self._check_XYZ(XYZ)
        x = x-x0
        y = y-y0
        z = z-z0
        r = np.sqrt(x**2 + y**2 + z**2)

        if field != 'total':
            Vp = -E0 * x
            if field == 'primary':
                return Vp

        Vt = np.zeros_like(r)
        ind0 = r > self.radius
        # total potential outside the sphere
        Vt[ind0] = -E0*x[ind0]*(1.-sig_cur*self.radius**3./r[ind0]**3.)
        # inside the sphere
        Vt[~ind0] = -E0*x[~ind0]*3.*sig0/(sig1+2.*sig0)

        if field == 'total':
            return Vt
        # field was not primary or total
        Vs = Vt - Vp
        if field == 'secondary':
            return Vs
        return Vt, Vp, Vs

    def electric_field(self, XYZ, field='all'):
        """Electric field for a sphere in a uniform wholespace

        Parameters
        ----------
        XYZ : (3, ) tuple of np.ndarray or (..., 3) np.ndarray
            locations to evaluate at. If a tuple, all
            the numpy arrays must be the same shape.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Et, Ep, Es : (..., 3) np.ndarray
            If field == "all"
        E : (..., 3) np.ndarray
            If only requesting a single field.
        """
        sig0 = self.sigma_background
        sig1 = self.sigma_sphere
        E0 = self.amplitude
        sig_cur = (sig1 - sig0) / (sig1 + 2 * sig0)

        x, y, z = self._check_XYZ(XYZ)
        x0, y0, z0 = self.location
        x = x-x0
        y = y-y0
        z = z-z0
        r = np.sqrt(x**2 + y**2 + z**2)

        if field != 'total':
            Ep = np.zeros((*x.shape, 3))
            Ep[..., 0] = E0
            if field == 'primary':
                return Ep

        Et = np.zeros((*x.shape, 3))
        ind0 = r > self.radius
        # total field outside the sphere
        Et[ind0, 0] = E0 + E0*self.radius**3./(r[ind0]**5.)*sig_cur*(2.*x[ind0]**2.-y[ind0]**2.-z[ind0]**2.)
        Et[ind0, 1] = E0*self.radius**3./(r[ind0]**5.)*3.*x[ind0]*y[ind0]*sig_cur
        Et[ind0, 2] = E0*self.radius**3./(r[ind0]**5.)*3.*x[ind0]*z[ind0]*sig_cur
        # inside the sphere
        Et[~ind0, 0] = 3.*sig0/(sig1+2.*sig0)*E0

        if field == 'total':
            return Et
        # field was not primary or total
        Es = Et - Ep
        if field == 'secondary':
            return Es
        return Et, Ep, Es

    def current_density(self, XYZ, field='all'):
        """Current density for a sphere in a uniform wholespace

        Parameters
        ----------
        XYZ : (3, ) tuple of np.ndarray or (..., 3) np.ndarray
            locations to evaluate at. If a tuple, all
            the numpy arrays must be the same shape.
        field : {'all', 'total', 'primary', 'secondary'}

        Returns
        -------
        Jt, Jp, Js : (..., 3) np.ndarray
            If field == "all"
        J : (..., 3) np.ndarray
            If only requesting a single field.
        """

        Et, Ep, Es = self.electric_field(XYZ, field='all')
        if field != 'total':
            Jp = self.sigma_background * Ep
            if field == 'primary':
                return Jp

        x, y, z = self._check_XYZ(XYZ)
        x0, y0, z0 = self.location
        x = x-x0
        y = y-y0
        z = z-z0
        r = np.sqrt(x**2 + y**2 + z**2)

        sigma = np.full(r.shape, self.sigma_background)
        sigma[r <= self.radius] = self.sigma_sphere

        Jt = sigma[..., None] * Et
        if field == 'total':
            return Jt

        Js = Jt - Jp
        if field == 'secondary':
            return Js
        return Jt, Jp, Js

    def charge_density(self, XYZ, dr=None):
        """charge density on the surface of a sphere in a uniform wholespace

        Parameters
        ----------
        XYZ : (3, ) tuple of np.ndarray or (..., 3) np.ndarray
            locations to evaluate at. If a tuple, all
            the numpy arrays must be the same shape.
        dr : float, optional
            Buffer around the edge of the sphere to calculate
            current density. Defaults to 5 % of the sphere radius

        Returns
        -------
        rho: (..., ) np.ndarray
        """

        sig0 = self.sigma_background
        sig1 = self.sigma_sphere
        sig_cur = (sig1 - sig0) / (sig1 + 2 * sig0)
        Ep = self.electric_field(XYZ, field='primary')

        x, y, z = self._check_XYZ(XYZ)
        x0, y0, z0 = self.location
        x = x-x0
        y = y-y0
        z = z-z0
        r = np.sqrt(x**2 + y**2 + z**2)

        if dr is None:
            dr = 0.05 * self.radius

        ind = (r < self.radius + 0.5*dr) & (r > self.radius - 0.5*dr)

        rho = np.zeros_like(r)
        rho[ind] = epsilon_0*3.*Ep[ind, 0]*sig_cur*x[ind]/(np.sqrt(x[ind]**2.+y[ind]**2.))

        return rho
